/**********************************************************************
   Mark Butcher    Bsc (Hons) MPhil MIET

   M.J.Butcher Consulting
   Birchstrasse 20f,    CH-5406, Rtihof
   Switzerland

   www.uTasker.com    Skype: M_J_Butcher

   ---------------------------------------------------------------------
   File:        debug.c
   Project:     uTasker Demonstration project
   ---------------------------------------------------------------------
   Copyright (C) M.J.Butcher Consulting 2004..2010
   *********************************************************************

   18.02.2007 Corrected port support for M52233 DEMO board               {1}
   26.02.2007 Improve port control via serial/Telnet                     {2}
   03.06.2007 Add FTP timeout and mode when configuring FTP server       {3}
   02.07.2007 Add LPC23XX GPIOs
   26.09.2007 Add LM3SXXXX GPIOs
   31.12.2007 Improve NE64 user port update                              {4}
   01.01.2008 ping test made conditional on ICMP_SEND_PING
   26.01.2008 parameter tests made conditional on parameter system       {5}
   02.05.2008 Add dummy code to allow the project to use serial interface without USE_MAINTENANCE {6}
   25.05.2008 Move fnAddTime() outside of USE_MAINTENANCE dependency     {7}
   09.08.2008 Add USB_INTERFACE debug interface support                  {8}
   08.10.2008 User POWER_UP macro for SAM7X code                         {9}
   05.02.2009 Remove some serial interface code when DEMO_UART not defined {10}
   26.04.2009 Add AVR32 support                                          {11}
   09.05.2009 Mask user ports when setting output                        {12}
   01.07.2009 Adapt for STRING_OPTIMISATION configuration (fnDebugDec() replaced by fnBufferDec() when used with string buffer), pointers used from uStrcpy no longer decremented
   24.07.2009 Add DO_RESET_ETHERNET_STATS support, which was previously missing {13}
   13.08.2009 Add simulator port call to ensure Luminary accuracy        {14}
   09.10.2009 Don't initialise port bit 1 when working with CGLCD (Luminary) {15}
   18.11.2009 Add AVR32_EVK1100 and AVR32_EVK1101 support                {16}
   28.11.2009 Add SD-card disk interface                                 {17}
   19.12.2009 Automatic check of TELNET output buffer when the TELNET connection is in use {18}
   31.01.2010 Remove "root" command since this can be achieved with "cd \" {19}
   31.01.2010 File open for rename uses attribute UTFAT_OPEN_FOR_RENAME rather than UTFAT_OPEN_FOR_WRITE {20}
   05.02.2010 Add IPV6                                                   {21}
   05.02.2010 Return value from fnGetARPentry() checked against 0 rather than against cucNullMACIP since the return value changed {22}
   21.05.2010 Add M5208 support                                          {23}
   13.06.2010 Change SD card state recognition to show unformatted information {24}

*/


/* =================================================================== */
/*                           include files                             */
/* =================================================================== */

#include "config.h"


/* =================================================================== */
/*                          local definitions                          */
/* =================================================================== */

//#define TEST_IIC_INTERFACE

#if defined USE_TELNET || defined SERIAL_INTERFACE || defined USB_INTERFACE // {8}
    #define KEEP_DEBUG
#endif

#if defined USE_MAINTENANCE || defined SERIAL_INTERFACE
    unsigned short usData_state = 0;
#endif
#if defined USE_MAINTENANCE || defined USB_INTERFACE                     // {8}
    #ifndef USE_MAINTENANCE
        unsigned short usUSB_state = ES_USB_RS232_MODE;
    #else
        unsigned short usUSB_state = 0;
    #endif
#endif

#ifdef USE_MAINTENANCE

#define OWN_TASK                TASK_DEBUG

#define TCP_SERVER_TEST         5

#define STALL_CRITICAL          (TX_BUFFER_SIZE/2)                       // when the Windows TCP implementation sees that the window is less than half the maximum windows size we have advertised it will perform a delay of about 5s.

#define IIC_WRITE_ADDRESS       0xd0
#define IIC_READ_ADDRESS        0xd1

#define UT_PATH_LENGTH          256                                      // length of maximum directory string

#define STALL_DIR_LISTING       1
#define STALL_PRINTING_FILE     2
#define STALL_COUNTING_CLUSTERS 3

                                                                         // we want to avoid this and so ensure that our TCP informs when the Winodw has opened again.
#ifdef KEEP_DEBUG
    #ifdef USE_TELNET
static int  fnTELNETListener(USOCKET Socket, unsigned char ucEvent, unsigned char *ucIp_Data, unsigned short usPortLen);
    #endif
static int  fnDisplayHelp(int iStart);

    #if defined USE_ICMP && defined ICMP_SEND_PING
static void fnPingTest(USOCKET dummy_socket);
static void fnPingSuccess(USOCKET dummy_socket);
static unsigned char ucTempIP[IPV4_LENGTH];
    #endif
    #if defined USE_IPV6 && defined ICMP_SEND_PING
static void fnPingIPV6Success(USOCKET dummy_socket);
static void fnPingV6Test(USOCKET dummy_socket);
static unsigned char ucPingTestV6[IPV6_LENGTH];
    #endif
static void fnDoTelnet(unsigned char ucType, CHAR *ptrInput);
    #ifdef SDCARD_SUPPORT                                                // {17}
        static int  fnDoDisk(unsigned char ucType, CHAR *ptrInput);
        static void fnSendPrompt(void);
    #endif
#endif
#ifdef TEST_CLIENT_SERVER
    static int fnTestTCPServer(int iAction);
        #define TCP_TEST_SERVER_START    0
        #define START_SERVER_TEST        1
        #define TCP_SERVER_TX            2
        #define RESET_SERVER_MESSAGE     3
        #define INCREMENT_SERVER_MESSAGE 4
        #define TCP_TEST_SERVER_LISTEN   5
        #define CONNECT_REMOTE           6
        #define TCP_SERVER_REPEAT        7
        #define ALLOW_NEXT_MESSAGE       8
#endif

#ifdef KEEP_DEBUG                                                        // if neither Telnet nor serial, we save space by removing debug task code

static USOCKET         Telnet_socket = -1;
static int             iMenuLocation = 0;
#if defined SERIAL_INTERFACE
    static unsigned char  ucPasswordState = PASSWORD_IDLE;
#endif
static unsigned short  usTelnet_state = 0;
static unsigned char   ucMenu = 0;
static unsigned char   ucDebugCnt = 0;
static unsigned char   ucTelnetMode;
#ifdef SDCARD_SUPPORT                                                    // {17}
    static UTDIRECTORY *ptr_utDirectory = 0;                             // pointer to a directory object
    static int iFATstalled = 0;                                          // stall flags when listing large directories and printing content
#endif

#define TCP_FLOW_CONTROL    1

#if defined IIC_INTERFACE && defined TEST_IIC_INTERFACE
    extern QUEUE_HANDLE IICPortID;
#endif

#define MAX_TAB_MARK              27                                     // this defines the maximum tab length supported
#define MAX_MENU_LINE             66                                     // a line of a menu should not be longer than this

#define DO_HELP                   0                                      // reference to help handler
    #define DO_MAIN_HELP            0                                    // specific help command to display main help menu
    #define DO_HELP_AT              1                                    // specific help command to display AT help menu
    #define DO_HELP_TELNET          2                                    // specific help command to display TELNET help menu

    #define DO_MENU_HELP_LAN        4                                    // go to LAN menu
    #define DO_MENU_HELP_SERIAL     5                                    // go to serial menu
    #define DO_MENU_HELP_IO         6                                    // go to I/O menu
    #define DO_MENU_HELP_ADMIN      7                                    // go to Administration menu
  //#define DO_MENU_HELP_SERVICES   8                                    // go to services menu
    #define DO_MENU_HELP_STATS      9                                    // go to statistics menu
    #define DO_MENU_HELP_USB        10                                   // go to USB menu
    #define DO_MENU_HELP_TCP        11                                   // go to TCP menu
  //#define DO_MENU_HELP_TCP_SER    12                                   // TCP/serial configs
    #define DO_MENU_HELP_PHY        12
    #define DO_MENU_HELP_OLED       13
    #define DO_MENU_HELP_I2C        14
    #define DO_MENU_HELP_DISK       15                                   // {17}
    #define DO_HELP_UP              16                                   // go up menu tree

#define DO_HARDWARE               1                                      // reference to hardware group of commands
    #define DO_RESET                0                                    // specific hardware command to reset target
    #define DO_DISPLAY_MEMORY_USE   1                                    // specific hardware command to display memory use
    #define DO_DDR                  2                                    // specific hardware command to set port direction
    #define DO_INPUT                4                                    // specific hardware command to read input port values
    #define DO_OUTPUT               5                                    // specific hardware command to set output port values
    #define DO_READ_ADC             6                                    // specific hardware command to read ADC inputs
    #define DO_READY_DOWNLOAD       7                                    // specific hardware command to prepare for a download
    #define DO_SET_USER_OUTPUT      8                                    // specific hardware command to configure user port outputs
    #define DO_GET_USER_OUTPUT      9                                    // specific hardware command to get user port outputs configuration
    #define DO_DISPLAY_UP_TIME     10                                    // specific hardware command to display up time
    #define DO_GET_DDR             11                                    // specific hardware command to get ddr
    #define DO_SAVE_PORT           12                                    // specific hardware command to save port settings

#define DO_TELNET                 2                                      // reference to Telnet group
    #define DO_TELNET_QUIT              0                                // specific Telnet comand to quit the session
    #define DO_TELNET_INACTIVITY_TIME   1                                // specific Telnet comand to set inactivity timeout
    #define DO_SET_TELNET_PORT          3                                // specific Telnet comand to set telnet port number
    #define DO_ENABLE_TELNET_DIALOUT    4                                // specific Telnet comand to enable/disable telnet dialout
    #define DO_ENABLE_TELNET_RAW_DIAL   5                                // specific Telnet comand to enable/disable telnet dialout in raw mode
    #define DO_ENABLE_TELNET_SCANNING   6                                // specific Telnet comand to enable/disable telnet escape sequence scanning
    #define DO_SET_TELNET_ESCAPE        7                                // specific Telnet comand to set escape sequence for command mode
    #define DO_RESUME_SESSION           8                                // specific Telnet comand to resume dial out data session
    #define DO_SERIAL_TCP_CONFIG        9                                // specific Telnet comand to show serial/TCP config

#define DO_FLASH                  3                                      // reference to FLASH group
    #define DO_SET_VALUE            0                                    // specific FLASH command to set a value
    #define DO_GET_VALUE            1                                    // specific FLASH command to get a value
    #define DO_SET_DEFAULTS         2                                    // specific FLASH command to return default parameters
    #define DO_INVALID_PARS         3                                    // specific FLASH command to invalidate all parameters
    #define DO_SET_TEMP_VALUE       4                                    // specific FLASH command to set a temporary value
    #define DO_VALIDATE_NEW_PARS    5                                    // specific FLASH command to validate new parameters
    #define DO_SAVE_PARS            6                                    // specific FLASH command to save and validate parameters
    #define DO_REJECT_CHANGES       7                                    // specific FLASH command to reset temporary modification

#define DO_IP                     4                                      // reference to IP group
    #define PING_TEST               0                                    // specific IP command to send ping test
    #define DO_SET_IP_ADDRESS       1                                    // specific IP command to set device's IPV4 address
    #define DO_SET_IP_SUBNET_MASK   2                                    // specific IP command to set device's subnet mask
    #define DO_SET_DEFAULT_GATEWAY  3                                    // specific IP command to set default gateway
    #define DO_SHOW_ETHERNET_STATS  4                                    // specific IP command to show ethernet statistics
    #define DO_SHOW_IP_CONFIG       5                                    // specific IP command to show IP configuration
    #define DO_SET_LAN_SPEED        6                                    // specific IP command to set speed of LAN
    #define SHOW_ARP                7                                    // specific IP command to show ARP table contents
    #define DELETE_ARP              8                                    // specific IP command to delete ARP table contents
    #define DO_SET_MAC              9                                    // specific IP command to set a new MAC address
    #define DO_SHOW_IP_CONFIG_TEMP  10                                   // specific IP command to show the temporary IP configuration
    #define DO_RESET_ETHERNET_STATS 11                                   // specific IP command to reset the ethernet statistics
    #define DO_SET_IPV6_ADDRESS     12                                   // specific IP command to set device's IPV6 address
    #define SHOW_NN                 13                                   // specific IP command to show IPV6 neighbor table contents
    #define DELETE_NN               14                                   // specific IP command to delete IPV6 neighbor table contents
    #define PING_TEST_V6            15                                   // specific IP command to send IPV6 ping test
    
#define DO_SERIAL                 5                                      // reference to serial group
    #define SERIAL_SET_BAUD               0                              // specific serial command to set baud rate
    #define SERIAL_SET_STOP               1                              // specific serial command to set stop bits
    #define SERIAL_SET_DATA_LENGTH        2                              // specific serial command to set data character length
    #define SERIAL_SET_PARITY             3                              // specific serial command to set parity
    #define SERIAL_SET_FLOW               4                              // specific serial command to set baud flow control
    #define DO_ENABLE_SERIAL_ETHERNET     5                              // specific serial command to enable/disable automatic serial->ethernet telnet connection
    #define DO_ENABLE_SERIAL_ESCAPE_SCAN  6                              // specific serial command to enable/disable escape sequence scanning
    #define DO_SET_SERIAL_ESCAPE          7                              // specific serial command to set escape sequence to be scanned
    #define SERIAL_SET_WAIT_BUFFER_SIZE   8                              // specific serial command to set wait buffer size
    #define SERIAL_SET_WAIT_BUFFER_TIME   9                              // specific serial command to set wait buffer wait time
    #define DO_ENABLE_ETHERNET_SERIAL     10                             // specific serial command to enable/disable automatic ethernet->serial connection
    #define RESERVED_COMMAND              11                             // specific serial command - reserved for the moment
    #define SERIAL_SET_DIALOUT_PORT       12                             // specific serial command to set dialout port number
    #define SERIAL_ENABLE_RAW_DIALOUT     13                             // specific serial command to enable/disable dial out in raw mode
    #define SERIAL_ENABLE_SERIAL_ETHERNET 14                             // specific serial command to enable/disable serial->ethernet dialin
    #define SERIAL_DIALIN_PORT            15                             // specific serial command to set serial->ethernet dialin port number
    #define SERIAL_DIALIN_IP_ADDRESS      16                             // specific serial command to set serial->ethernet dialin IP address
    #define SERIAL_SHOW_CONFIG            17                             // specific serial command to show serial configuration
    #define SERIAL_SHOW_STATS             18                             // specific serial command to show serial statistics
    #define SERIAL_TRUSTED_IP             19                             // specific serial command to set a trusted IP address
    #define DO_ENABLE_BINARY_CONNECTION   20                             // specific serial command to set binary / Telnet data connection
    #define SERIAL_SET_HIGH_WATER         21                             // specific serial command to set flow control high water level
    #define SERIAL_SET_LOW_WATER          22                             // specific serial command to set flow control low water level
    #define SERIAL_ENABLE_AUTO_SER_ETH    23                             // specific serial command to enable/disable auto dial-in
    #define SERIAL_ENABLE_AUTO_SER_ETH_HW 24                             // specific serial command to enable/disable DTR rather than data dial-in (when auto dial-in enabled)
    #define RESET_SERIAL_STATS            25                             // specific serial command to reset serial statistics
    #define DO_ENABLE_IAC_TX              26                             // specific serial command to enable/disable IAC stuffing when in RAW mode
    #define DO_ENABLE_IAC_RX              27                             // specific serial command to enable/disable IAC filtering when in RAW mode

#define DO_ADMIN                  6                                      // reference to administration group
    #define DO_SETNAME_PASS         0                                    // specific administration command to set user and password
    #define DO_SHOW_STATS           1                                    // specific administration command to show statistics
    #define DO_SHOW_CONFIG          2                                    // specific administration command to show configuration
    #define DO_SHOW_INFO            3                                    // specific administration command to show information
    #define DO_SET_DEVICE_NAME      4                                    // specific administration command to set a device identification name
    #define DO_SET_SERIAL_NUMBER    5                                    // specific administration command to set new serial number
    #define DO_CONNECT              6                                    // specific administration command to connect to test server
    #define DO_START                7                                    // specific administration command to start server test

#define DO_SERVER                 7                                      // reference to server group
    #define DO_ENABLE_TELNET        0                                    // specific server command to enable/disable telnet
    #define DO_ENABLE_DHCP          1                                    // specific server command to enable/disable DHCP client
    #define DO_ENABLE_FTP           2                                    // specific server command to enable/disable FTP server
    #define DO_ENABLE_WEB_SERVER    3                                    // specific server command to enable/disable HTTP server
    #define SHOW_SECURITY           4                                    // specific server command to show security settings
    #define DO_HTTP_AUTH            5                                    // specific server command to enable/disable HTTP authentication
    #define DO_ENABLE_LOGIN         6                                    // specific server command to enable/disable terminal menu login

#define DO_USB                    8                                      // reference to USB group
    #define DO_USB_RS232_MODE       0                                    // specific USB command to go to RS232-USB mode
    #define DO_USB_DOWNLOAD         1                                    // specific USB command to download firmware
    #define DO_USB_UPLOAD           2                                    // specific USB command to upload data

#define DO_OLED                   9                                      // reference to OLED group
    #define DO_OLED_GET_GRAY_SCALES 0                                    // specific OLED command to get gray levels
    #define DO_OLED_SET_GRAY_SCALE  1                                    // specific OLED command to set gray level value

#define DO_I2C                    10                                     // reference to I2C group
    #define DO_I2C_WRITE            0                                    // specific I2C command to write to I2C bus
    #define DO_I2C_READ             1                                    // specific I2C command to read from I2C bus
    #define DO_I2C_READ_NO_ADDRESS  2                                    // specific I2C command to read from I2C bus without first sending an address
    #define DO_I2C_READ_PLUS_WRITE  3                                    // specific I2C command to read from I2C bus frompresent address, followed by a queued write to set address zero

#define DO_DISK                   11                                     // reference to SD-card, disk group
    #define DO_DIR                  0                                    // specific command to display directory content
    #define DO_CHANGE_DIR           1                                    // specific command to change directory
    #define DO_NEW_FILE             2                                    // specific command to create a file
    #define DO_NEW_DIR              3                                    // specific command to create a directory
    #define DO_WRITE_FILE           4                                    // specific command to add file content
    #define DO_PRINT_FILE           5                                    // specific command to print the content of a file
    #define DO_ROOT                 6                                    // specific command to set root directory
    #define DO_DELETE               7                                    // specific command to delete a file or directory
    #define DO_DISPLAY_SECTOR       8                                    // specific command to display a sector of the SD card
    #define DO_FORMAT               9                                    // specific command to format a disk
    #define DO_REFORMAT             10                                   // specific command to reformat a disk
    #define DO_INFO                 11                                   // specific command to display information about SD card and utFAT
    #define DO_RENAME               12                                   // specific command to rename file or directory

#define DO_TEST                   12                                     // reference to test group
    #define DO_START_TEST_1         0                                    // specific server command to start test 1
    #define DO_DUMP_PHY             1                                    // specific command to dump PHY register contents

#define MENU_HELP_LAN             1
#define MENU_HELP_SERIAL          2
#define MENU_HELP_IO              3
#define MENU_HELP_ADMIN           4
#define MENU_HELP_STATS           5
#define MENU_HELP_USB             6
#define MENU_HELP_I2C             7
#define MENU_HELP_DISK            8

/* =================================================================== */
/*                      local structure definitions                    */
/* =================================================================== */

typedef struct stDEBUG_COMMAND
{
    CHAR    *strCommand;                                                 // the command string to accept
    CHAR    *strText;                                                    // the help text related to this command
    unsigned char  ucDoFunction;                                         // reference to the function handling the command
    unsigned char  ucType;                                               // the command type
} DEBUG_COMMAND;

typedef struct stMENUS
{
    const DEBUG_COMMAND  *pMenu;                                         // pointer to the menu
    unsigned char  ucTabSize;                                            // the tab width for this menu
    unsigned char  ucMenuLength;                                         // the length of this menu
    CHAR           *cTitle;                                              // menu caption
} MENUS;


/* =================================================================== */
/*                             constants                               */
/* =================================================================== */

#ifdef USE_TELNET
    static const UTASK_TASK OurTask = OWN_TASK;
#endif

static const CHAR cDeleteInput[] = {DELETE_KEY, ' ', DELETE_KEY, 0};

static const DEBUG_COMMAND tMainCommand[] = {
    {"1",                 "Configure LAN interface",               DO_HELP,          DO_MENU_HELP_LAN },
    {"2",                 "Configure serial interface",            DO_HELP,          DO_MENU_HELP_SERIAL },
    {"3",                 "Go to I/O menu",                        DO_HELP,          DO_MENU_HELP_IO },
    {"4",                 "Go to administration menu",             DO_HELP,          DO_MENU_HELP_ADMIN },
    {"5",                 "Go to overview/statistics menu",        DO_HELP,          DO_MENU_HELP_STATS },
    {"6",                 "Go to USB menu",                        DO_HELP,          DO_MENU_HELP_USB },
    {"7",                 "Go to I2C menu",                        DO_HELP,          DO_MENU_HELP_I2C },
    {"8",                 "Go to SD-Card disk interface",          DO_HELP,          DO_MENU_HELP_DISK }, // {17}
    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};

static const DEBUG_COMMAND tLANCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },

    {"set_dhcp",          "<enable/disable> DHCP",                 DO_SERVER,        DO_ENABLE_DHCP },
    {"set_ip_add",        "Set IPV4 address",                      DO_IP,            DO_SET_IP_ADDRESS },
#ifdef USE_IPV6                                                          // {21}
    {"set_ipv6_add",      "Set IPV6 address",                      DO_IP,            DO_SET_IPV6_ADDRESS },
#endif
    {"set_ip_mask",       "Set IP subnet mask",                    DO_IP,            DO_SET_IP_SUBNET_MASK },
    {"set_ip_gway",       "Set default gateway",                   DO_IP,            DO_SET_DEFAULT_GATEWAY },
    {"set_eth_speed",     "Set LAN speed (10/100/AUTO)",           DO_IP,            DO_SET_LAN_SPEED },
    {"show_config",       "Show network configuration",            DO_IP,            DO_SHOW_IP_CONFIG },
    {"show_config_o",     "Show original network configuration",   DO_IP,            DO_SHOW_IP_CONFIG_TEMP },
#ifdef USE_PARAMETER_BLOCK                                           // {5}
    {"save",              "Save configuration to FLASH",           DO_FLASH,         DO_SAVE_PARS },
    {"validate",          "Validate temporary config. in FLASH",   DO_FLASH,         DO_VALIDATE_NEW_PARS },
#endif
#if defined USE_ICMP && defined ICMP_SEND_PING
    {"ping",              "Ping test IP address",                  DO_IP,            PING_TEST },
#endif
    {"arp -d",            "Delete ARP table",                      DO_IP,            DELETE_ARP },
#if defined USE_IPV6 && defined ICMP_SEND_PING                           // {21}
    {"pingv6",            "Ping test IPV6 address",                DO_IP,            PING_TEST_V6 },
#endif
    {"arp -a",            "Request ARP table",                     DO_IP,            SHOW_ARP },
    {"arp -d",            "Delete ARP table",                      DO_IP,            DELETE_ARP },
#ifdef USE_IPV6                                                          // {21}
    {"nn -a",             "Request IPV6 neighbor table",           DO_IP,            SHOW_NN },
    {"nn -d",             "Delete IPV6 neighbor table",            DO_IP,            DELETE_NN },
#endif

    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};

#ifdef SERIAL_INTERFACE
static const DEBUG_COMMAND tSERIALCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },

    {"set_baud",          "Set serial baud rate",                  DO_SERIAL,        SERIAL_SET_BAUD },
    {"set_stop",          "Set stop bits (1/1.5/2)",               DO_SERIAL,        SERIAL_SET_STOP },
    {"set_bits",          "Set data bits (7/8)",                   DO_SERIAL,        SERIAL_SET_DATA_LENGTH },
    {"set_par",           "Set parity (EVEN/ODD/NONE)",            DO_SERIAL,        SERIAL_SET_PARITY },
    {"set_flow",          "Set flow control (XON/RTS/NONE)",       DO_SERIAL,        SERIAL_SET_FLOW },
    {"set_high_water",    "Set flow stall (%) [1..99]",            DO_SERIAL,        SERIAL_SET_HIGH_WATER },
    {"set_low_water",     "Set flow restart (%) [1..99]",          DO_SERIAL,        SERIAL_SET_LOW_WATER },
    {"show_config",       "Show serial configuration",             DO_SERIAL,        SERIAL_SHOW_CONFIG },
    #ifdef USE_PARAMETER_BLOCK                                           // {5}
    {"save",              "Save configuration to FLASH",           DO_FLASH,         DO_SAVE_PARS },
    #endif

    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};
#else
static const DEBUG_COMMAND tSERIALCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },
    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};
#endif

static const DEBUG_COMMAND tIOCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },

//  {"set_user",          "Set output mode [0..3] [<d>|<u>]",      DO_HARDWARE,      DO_SET_USER_OUTPUT },
//  {"get_user",          "Get output mode [0..3]",                DO_HARDWARE,      DO_GET_USER_OUTPUT },
    {"set_ddr",           "Set port type [1..4] [<i>|<o>",         DO_HARDWARE,      DO_DDR },
    {"get_ddr",           "Get data direction [1..4]",             DO_HARDWARE,      DO_GET_DDR },
    {"read_port",         "Read port input [1..4]",                DO_HARDWARE,      DO_INPUT },
    {"write_port",        "Set port output [1..4] [0/1]",          DO_HARDWARE,      DO_OUTPUT },
    {"save",              "Save port setting as default",          DO_HARDWARE,      DO_SAVE_PORT },

    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};

static const DEBUG_COMMAND tADMINCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },

    {"set_ftp",           "<enable/disable> FTP server",           DO_SERVER,        DO_ENABLE_FTP },
    {"set_telnet",        "<enable/disable> Telnet service",       DO_SERVER,        DO_ENABLE_TELNET },
    {"set_web",           "<enable/disable> WEB server",           DO_SERVER,        DO_ENABLE_WEB_SERVER },
    {"set_web_auth",      "<enable/disable> WEB server authentication", DO_SERVER,   DO_HTTP_AUTH },
    {"show_config",       "Show configuration",                    DO_ADMIN,         DO_SHOW_CONFIG },
#ifdef USE_PARAMETER_BLOCK                                           // {5}
    {"save",              "Save configuration to FLASH",           DO_FLASH,         DO_SAVE_PARS },
    {"reject",            "Reset non-saved changes",               DO_FLASH,         DO_REJECT_CHANGES },
    {"restore",           "Restore factory settings",              DO_FLASH,         DO_SET_DEFAULTS },
#endif
#ifdef TEST_CLIENT_SERVER
    {"connect",           "Connect to server",                     DO_ADMIN,         DO_CONNECT },
    {"test",              "Start server test",                     DO_ADMIN,         DO_START },
#endif
    {"reset",             "Reset device",                          DO_HARDWARE,      DO_RESET },

    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};

static const DEBUG_COMMAND tStatCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },

    {"ipstat",            "Show Ethernet statistics",              DO_IP,            DO_SHOW_ETHERNET_STATS },
    {"r_ipstat",          "Reset Ethernet statistics",             DO_IP,            DO_RESET_ETHERNET_STATS },
    {"up_time",           "Show operating time",                   DO_HARDWARE,      DO_DISPLAY_UP_TIME },
    {"memory",            "Show memory use",                       DO_HARDWARE,      DO_DISPLAY_MEMORY_USE },

    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};

static const DEBUG_COMMAND tUSBCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },
#ifdef USB_INTERFACE
    {"usb-serial",        "RS232<->USB mode (disconnect to quit)", DO_USB,           DO_USB_RS232_MODE },
    {"usb-load",          "USB-SW download",                       DO_USB,           DO_USB_DOWNLOAD },   
#endif
    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};

static const DEBUG_COMMAND tI2CCommand[] = {
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },
#if defined IIC_INTERFACE && defined TEST_IIC_INTERFACE
    {"wr",                "write [add] [value]",                   DO_I2C,           DO_I2C_WRITE },
    {"rd",                "read  [add] [number of bytes]",         DO_I2C,           DO_I2C_READ },
    {"srd",               "simple read [number of bytes]",         DO_I2C,           DO_I2C_READ_NO_ADDRESS },
    {"rdq",               "read  [add] [number of bytes] + wr",    DO_I2C,           DO_I2C_READ_PLUS_WRITE },
#endif
    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};

static const DEBUG_COMMAND tDiskCommand[] = {                            // {17}
    {"up",                "go to main menu",                       DO_HELP,          DO_HELP_UP },
#if defined SDCARD_SUPPORT
    {"info",              "utFAT/card info",                       DO_DISK,          DO_INFO },
    {"dir",               "[path] show directory content",         DO_DISK,          DO_DIR },
    {"cd",                "[path] change dir. (.. for up)",        DO_DISK,          DO_CHANGE_DIR },    
    {"file" ,             "[path] new empty file",                 DO_DISK,          DO_NEW_FILE },
    {"write",             "[path] test write to file",             DO_DISK,          DO_WRITE_FILE },
    {"mkdir" ,            "new empty dir",                         DO_DISK,          DO_NEW_DIR },
    {"rename" ,           "[from] [to] rename",                    DO_DISK,          DO_RENAME },
    {"print",             "[path] print file content",             DO_DISK,          DO_PRINT_FILE },
  //{"root",              "set root dir",                          DO_DISK,          DO_ROOT }, {19}
    {"del",               "[path] delete file or dir.",            DO_DISK,          DO_DELETE },
    {"format",            "[label] format (unformatted) disk",     DO_DISK,          DO_FORMAT },
    {"re-format",         "[label] reformat disk!!!!!",            DO_DISK,          DO_REFORMAT },
    {"sect",              "[hex number] display sector",           DO_DISK,          DO_DISPLAY_SECTOR },    
#endif
    {"help",              "Display menu specific help",            DO_HELP,          DO_MAIN_HELP },
    {"quit",              "Leave command mode",                    DO_TELNET,        DO_TELNET_QUIT },
};


// Special secret menu - not displayed as menu...
//
static const DEBUG_COMMAND tSecretCommands[] = {
    {"MAC",               0,                 DO_IP,            DO_SET_MAC },
};

static const MENUS ucMenus[] = {
    { tMainCommand,     15, (sizeof(tMainCommand) / sizeof(DEBUG_COMMAND)),   "     Main menu"},
    { tLANCommand,      20, (sizeof(tLANCommand) / sizeof(DEBUG_COMMAND)),    " LAN configuration"},
    { tSERIALCommand,   22, (sizeof(tSERIALCommand) / sizeof(DEBUG_COMMAND)), "   Serial config."},
    { tIOCommand,       15, (sizeof(tIOCommand) / sizeof(DEBUG_COMMAND)),     " Input/Output menu"},
    { tADMINCommand,    17, (sizeof(tADMINCommand) / sizeof(DEBUG_COMMAND)),  "   Admin. menu"},
    { tStatCommand,     17, (sizeof(tStatCommand) / sizeof(DEBUG_COMMAND)),   "   Stats. menu"},
    { tUSBCommand,      13, (sizeof(tUSBCommand) / sizeof(DEBUG_COMMAND)),    "    USB menu"},
    { tI2CCommand,      13, (sizeof(tI2CCommand) / sizeof(DEBUG_COMMAND)),    "    I2C menu"},
    { tDiskCommand,     13, (sizeof(tDiskCommand) / sizeof(DEBUG_COMMAND)),   "  Disk interface"}, // {17}
};






// Debug task
//
extern void fnDebug(TTASKTABLE *ptrTaskTable)
{
    QUEUE_HANDLE        PortIDInternal = ptrTaskTable->TaskID;           // queue ID for task input
    unsigned char       ucInputMessage[SMALL_MESSAGE];                   // reserve space for receiving messages

    if (Telnet_socket < 0) {
#ifdef USE_TELNET
        fnConfigureTelnetServer();
#endif
    }

    while (fnRead(PortIDInternal, ucInputMessage, HEADER_LENGTH)) {      // check input queue
        switch (ucInputMessage[MSG_SOURCE_TASK]) {                       // switch depending on message source
        case INTERRUPT_EVENT:
            if (TX_FREE == ucInputMessage[MSG_INTERRUPT_EVENT]) {
#ifdef SDCARD_SUPPORT                                                    // {17}
                if (iFATstalled != 0) {
                    if (iFATstalled == STALL_DIR_LISTING) {
                        fnDoDisk(DO_DIR, 0);                             // continue listing a directory content
                    }
                    else if (iFATstalled == STALL_PRINTING_FILE) {
                        fnDoDisk(DO_PRINT_FILE, 0);
                    }
                }
#endif
                if (iMenuLocation) {
                    fnDisplayHelp(iMenuLocation);
                }
                else if (ucTelnetMode & TCP_FLOW_CONTROL) {
                    ucTelnetMode &= ~TCP_FLOW_CONTROL;
#ifdef CONTROL_WINDOW_SIZE
                    fnReportWindow(Telnet_socket, TX_BUFFER_SIZE);
#endif
                }
            }
#ifdef SDCARD_SUPPORT
            else if (UTFAT_OPERATION_COMPLETED == ucInputMessage[MSG_INTERRUPT_EVENT]) {
                if (iFATstalled == STALL_COUNTING_CLUSTERS) {
                    UTDISK *ptrDiskInfo = fnGetDiskInfo(DISK_D);
                    if (ptrDiskInfo->utFileInfo.ulFreeClusterCount != 0xffffffff) {
                        fnDebugDec(((ptrDiskInfo->utFileInfo.ulFreeClusterCount * ptrDiskInfo->utFAT.ucSectorsPerCluster * ptrDiskInfo->utFAT.usBytesPerSector)), 0);
                        fnDebugMsg(" bytes free\r\n");
                        fnSendPrompt();
                    }
                }
            }
#endif
#ifdef TEST_CLIENT_SERVER
            else if (TCP_SERVER_TEST == ucInputMessage[MSG_INTERRUPT_EVENT]) {
                fnTestTCPServer(TCP_SERVER_TX);                          // send next test frame
            }
#endif
            break;

        case TASK_ARP:
            fnRead( PortIDInternal, ucInputMessage, ucInputMessage[MSG_CONTENT_LENGTH]);  // read the contents
            if (ucInputMessage[ 0 ] == ARP_RESOLUTION_SUCCESS) {
                // ARP has just resolved the address of our maintenance partner so we can resend the last message
                // This will not generally happen for server applications since we have just received a frame from the destination before sending ours
                fnDebugMsg("\n\rARP resolved\n\r");
                if ((USOCKET)ucInputMessage[1] < 0) {
#if defined USE_IP && defined USE_IPV6
                    if ((USOCKET)ucInputMessage[1] == -1) {              // single IPV6 ping test being performed to a destination outside of the local subnet - the IPV4 gatewas has been resolved
    #if defined ICMP_SEND_PING
                        fnPingV6Test((USOCKET)ucInputMessage[1]);        // repeat IPV6 ping test
    #endif
                    }
    #if defined USE_ICMP && defined ICMP_SEND_PING
                    else {
                        fnPingTest((USOCKET)ucInputMessage[1]);          // non-socket - it must be for our ping test
                    }
    #endif
#else
    #if defined USE_ICMP && defined ICMP_SEND_PING
                    fnPingTest((USOCKET)ucInputMessage[1]);              // non-socket - it must be for our ping test
    #endif
#endif
                }            
#ifdef USE_TELNET
                else {
                    fnSendBufTCP(Telnet_socket, 0, 0, TCP_BUF_REP);      // must be for telnet
                }
#endif
            }
            else if (ucInputMessage[0] == ARP_RESOLUTION_FAILED) {
                fnDebugMsg("\n\rARP resolution failed\n\r");             // we have learned that the partner MAC address could not be resolved...we ignore it in this case
            }
#ifdef USE_IPV6
            else if (ucInputMessage[0] == NN_RESOLUTION_SUCCESS) {
                fnDebugMsg("\n\rNN resolved\n\r");
                if ((USOCKET)ucInputMessage[1] < 0) {
    #if defined ICMP_SEND_PING
                    fnPingV6Test((USOCKET)ucInputMessage[1]);            // non-socket - it must be for our IPV6 ping test
    #endif
                }
            }
            else if (ucInputMessage[0] == NN_RESOLUTION_FAILED) {
                fnDebugMsg("\n\rNN resolution failed\n\r");              // we have learned that the partner MAC address could not be resolved...we ignore it in this case
            }
#endif
            break;

#if (defined USE_ICMP || defined USE_IPV6) && defined ICMP_SEND_PING
        case TASK_ICMP:
            fnRead( PortIDInternal, ucInputMessage, ucInputMessage[MSG_CONTENT_LENGTH]); // read the contents
    #if defined USE_ICMP && defined ICMP_SEND_PING
            if (ucInputMessage[0] == PING_RESULT) {
                fnPingSuccess((USOCKET)ucInputMessage[ 1 ]);
                break;
            }
    #endif
    #ifdef USE_IPV6
            if (ucInputMessage[0] == PING_IPV6_RESULT) {
                fnPingIPV6Success((USOCKET)ucInputMessage[1]);
            }
    #endif
            break;
#endif

        default:
            fnRead( PortIDInternal, ucInputMessage, ucInputMessage[MSG_CONTENT_LENGTH]); // flush any unexpected messages (assuming they arrived from another task)
            break;
        }
    }
#if defined IIC_INTERFACE && defined TEST_IIC_INTERFACE
    if (fnMsgs(IICPortID)) {                                             // if IIC message waiting
        QUEUE_TRANSFER iic_length = fnMsgs(IICPortID);
        int x = 0;
        fnDebugMsg("I2C Input = ");
        fnRead( IICPortID, ucInputMessage, iic_length);
        while (iic_length--) {
            fnDebugHex(ucInputMessage[x++], (WITH_LEADIN | WITH_SPACE | 1));// display received bytes
        }
        fnDebugMsg("\r\n");
    }
#endif
}


static void fnDisplaySerialNumber(void)
{
    CHAR cSN[18];                                                        // space for largest possible serial number with line feeds and null-terminator

    fnDebugMsg("\n\rSerial number: ");
    fnShowSN(cSN);
    fnDebugMsg(cSN);
}


static void fnDisplaySoftwareVersion(void)
{
    fnDebugMsg("Software version ");
    fnDebugMsg((CHAR *)cSoftwareVersion);
    fnDebugMsg("\n\r");
}

static void fnDisplayDeviceName(void)
{
    fnDebugMsg("Device identification: ");
    fnDebugMsg(temp_pars->temp_parameters.cDeviceIDName);
}



// check whether the command input is correct - if it is, return pointer to rest of input (skipping white space before possible parameters)
//
static CHAR *fnCheckInput(CHAR *str1, CHAR *str2)
{
    while (*str2) {
        if (*str1++ != *str2++) return 0;                                // no match
    }
    if (*str1 > ' ') return 0;                                           // we don't accept the match because too short

    while (*str1 == ' ') str1++;                                         // jump over trailing white space

    return str1;                                                         // return pointer to possible parameter
}

static int fnDisplayHelp(int iStart)
{
    const DEBUG_COMMAND *ptrCom = ucMenus[ucMenu].pMenu;
    unsigned char ucTab = ucMenus[ucMenu].ucTabSize;
    int iEntries = ucMenus[ucMenu].ucMenuLength;
    int iDisplayLen;
    unsigned char ucTabs[MAX_TAB_MARK];

    uMemset(ucTabs, ' ', MAX_TAB_MARK);

    if (iStart) {
        while (iEntries > iStart) {
          ptrCom++;
          iEntries--;
        }
    }
    else {
        fnDebugMsg("\n\n\r");
        fnDebugMsg(ucMenus[ucMenu].cTitle);
        fnDebugMsg("\n\r===================\n\r");
    }

    while (iEntries--) {
        if (!fnWrite(DebugHandle, 0, MAX_MENU_LINE)) {                   // the transmit buffer cannot accept more data so we will pause until it can
            fnDriver(DebugHandle, MODIFY_WAKEUP, (MODIFY_TX | OWN_TASK));// we want to be woken when the queue is free again
            iMenuLocation = (iEntries + 1);                              // save the position we stopped at
            return 1;
        }
        iDisplayLen = fnDebugMsg(ptrCom->strCommand);
        if (iDisplayLen >= ucTab) {
            fnWrite(DebugHandle, ucTabs, 1);
        }
        else {
            fnWrite(DebugHandle, ucTabs, (QUEUE_TRANSFER)(ucTab - iDisplayLen));
        }
        iDisplayLen = fnDebugMsg(ptrCom->strText);
        fnDebugMsg("\n\r");
        ptrCom++;
    }
    iMenuLocation = 0;
    return 0;
}

static void fnDisplaySystemMemory(void)
{
    fnDebugMsg("\n\rSystem memory use:\n\r");
    fnDebugMsg(    "==================\n\r");
    fnDebugMsg("Free heap = ");
    fnDebugHex(fnHeapFree(), (2 | WITH_LEADIN));
    fnDebugMsg(" from ");
    fnDebugHex(fnHeapAvailable(), (2 | WITH_LEADIN));
    fnDebugMsg("\n\rUnused stack = ");
    fnDebugHex(fnStackFree(), (2 | WITH_LEADIN));
    fnDebugMsg("\n\r");
}
#ifdef USE_PARAMETER_BLOCK                                               // {5}
static unsigned short fnGetWordShift(CHAR **ptrIn)
{
    unsigned short usResult = 0;
    unsigned char ucValue;
    CHAR *ptrInput = *ptrIn;
    int i = 4;

    while (*ptrInput == ' ') {                                           // jump over white space
       ++(*ptrIn);
       ptrInput++;
    }

    while ((i--) && (*ptrInput >= '0')) {
        usResult <<= 4;
        ucValue = *ptrInput++;
        ucValue -= '0';
        if (ucValue > 9) ucValue -= ('A' - '9' - 1);
        ucValue &= 0x0f;                                                 // makes previous line insensitive to case
        usResult |= ucValue;
        ++(*ptrIn);
    }

    return usResult;
}

static void fnDoFlash(unsigned char ucType, CHAR *ptrInput)
{
    unsigned short usParameter;
    unsigned char  ucValue;

    switch (ucType) {
    case DO_SET_VALUE:                                                   // set a parameter value
        usParameter = fnGetWordShift(&ptrInput);
        ucValue = (unsigned char)fnGetWordShift(&ptrInput);
        fnSetPar(usParameter, &ucValue, sizeof(ucValue));                // set parameter value to FLASH
        fnDebugMsg("\n\rParameter set\n\r");
        break;
    #ifdef USE_PAR_SWAP_BLOCK
    case DO_SET_TEMP_VALUE:
        usParameter = fnGetWordShift(&ptrInput);
        ucValue = (unsigned char)fnGetWordShift(&ptrInput);
        fnSetPar((unsigned short)(usParameter | TEMPORARY_PARAM_SET), (unsigned char *)&ucValue, sizeof(ucValue)); // set temporary parameter value to FLASH
        fnDebugMsg("\n\rTemp. Parameter set\n\r");
        break;
    #endif
    case DO_GET_VALUE:                                                   // get a parameter value from FLASH
        usParameter = fnGetWordShift(&ptrInput);
        if ((fnGetPar(usParameter, &ucValue, sizeof(ucValue)) >= 0)) {   // get parameter value from FLASH
            fnDebugMsg("\n\rParameter value = ");
            fnDebugHex(ucValue, (WITH_CR_LF | sizeof(ucValue)));
        }
        else {
            fnDebugMsg("\n\rNo valid parameters available\n\r");
        }
        break;
    case DO_REJECT_CHANGES:
        uMemcpy(&temp_pars->temp_network, &network, sizeof(NETWORK_PARAMETERS)); // reverse all changes
        uMemcpy(&temp_pars->temp_parameters, parameters, sizeof(PARS));
        fnDebugMsg("\n\rOriginal parameters restored\n\r");
        break;

    case DO_SET_DEFAULTS:                                                // delete all parameter values (reset to 0xff)
        fnRestoreFactory();
        fnSaveNewPars(SAVE_NEW_PARAMETERS);
        fnDebugMsg("\n\rFactory parameters restored\n\r");
        break;

    case DO_INVALID_PARS:                                                // delete and invalidate all parameter values
    #ifdef ACTIVE_FILE_SYSTEM
        fnDelPar(INVALIDATE_PARAMETER_BLOCK);
    #endif
        fnDebugMsg("\n\rParameters invalidated\n\r");
        break;
    #ifdef USE_PAR_SWAP_BLOCK
    case DO_VALIDATE_NEW_PARS:                                           // delete and invalidate present invalidate and switch to new parameter block
        #ifdef ACTIVE_FILE_SYSTEM
        fnDelPar(SWAP_PARAMETER_BLOCK);
        #endif
        fnDebugMsg("\n\rNew Parameters validated\n\r");
        break;
    #endif
    case DO_SAVE_PARS:                                                   // copy present parameters to FLASH and validate them
        fnSaveNewPars(SAVE_NEW_PARAMETERS);
        fnDebugMsg("\n\rConfiguration saved\n\r");
        break;
    }
}
#endif

#ifdef USE_IP
static void fnDisplayMAC(unsigned char *ucMac)
{
    CHAR cBuf[18];

    fnMACStr(ucMac, cBuf);
    fnDebugMsg(cBuf);
}

static void fnDisplayIP(unsigned char *ptrIP)
{
    CHAR cBuf[16];

    fnIPStr(ptrIP, cBuf);
    fnDebugMsg(cBuf);
}
#endif
#ifdef USE_IPV6                                                          // {21}
static void fnDisplayIPV6(unsigned char *ptrIP)
{
    CHAR cBuf[MAX_IPV6_STRING];

    fnIPV6Str(ptrIP, cBuf);
    fnDebugMsg(cBuf);
}
#endif

#if defined USE_ICMP && defined ICMP_SEND_PING
static void fnPingSuccess(USOCKET dummy_socket)
{
    unsigned char *ping_address;

    if (dummy_socket == -2) {
        ping_address = ucTempIP;
    }
    else {
        fnDebugMsg("Unknown PING response\r\n");
        return;
    }
    fnDebugMsg("PING response from ");
    fnDisplayIP(ping_address);
    fnDebugMsg("\r\n");
}

static void fnPingTest(USOCKET dummy_socket)
{
    unsigned char *ping_address;

    if (dummy_socket == -2) {
        ping_address = ucTempIP;
    }
    else {
        return;
    }
    if (NO_ARP_ENTRY == fnSendPing(ping_address, MAX_TTL, OWN_TASK, dummy_socket)) {
        fnDebugMsg("ARP resolution started\n\r");
    }
}
#endif

#if defined USE_IPV6 && defined ICMP_SEND_PING
static void fnPingIPV6Success(USOCKET dummy_socket)
{
    unsigned char *ping_address;

    if (dummy_socket == -1) {
        ping_address = ucPingTestV6;
    }
    else {
        fnDebugMsg("Unknown IPV6 PING response\r\n");
        return;
    }
    fnDebugMsg("IPV6 PING response from ");
    fnDisplayIPV6(ping_address);
    fnDebugMsg("\r\n");
}

static void fnPingV6Test(USOCKET dummy_socket)
{
    unsigned char *ping_IPV6_address;

    if (dummy_socket == -1) {
        ping_IPV6_address = ucPingTestV6;
    }
    else {
        return;
    }
    if (NO_ARP_ENTRY == fnSendV6Ping(ping_IPV6_address, MAX_TTL, OWN_TASK, dummy_socket)) {
        fnDebugMsg("NN resolution started\n\r");
    }
}
#endif

#if defined SERIAL_INTERFACE && defined DEMO_UART
#define NUMBER_OF_SPEEDS 12
static const CHAR cSpeeds[NUMBER_OF_SPEEDS][7] = {
    {"300"},
    {"600"},
    {"1200"},
    {"2400"},
    {"4800"},
    {"9600"},
    {"14400"},
    {"19200"},
    {"38400"},
    {"57600"},
    {"115200"},
    {"230400"},
};

static void fnShowSpeed(void)
{
    if (temp_pars->temp_parameters.ucSerialSpeed < NUMBER_OF_SPEEDS) {
        fnDebugMsg((CHAR *)cSpeeds[temp_pars->temp_parameters.ucSerialSpeed]);
    }
    else {
        fnDebugMsg("invalid!");
    }
}
#endif

#if defined SERIAL_INTERFACE && defined DEMO_UART
static int fnGetSpeed(CHAR *ptr_input, unsigned char *ucNew_speed)
{
    int i = 0;
    while (i < NUMBER_OF_SPEEDS) {
        if (!(uStrcmp((CHAR *)cSpeeds[i], ptr_input))) {
            *ucNew_speed = i;                                            // set corresponding speed index
            return 1;                                                    // return found
        }
        i++;
    }
    return 0;
}


#define NUMBER_OF_STOPS 3
static const CHAR cStops[NUMBER_OF_STOPS][4] = {
    {"1"},
    {"1.5"},
    {"2"},
};

static void fnShowStops(void)
{
    int i = 0;
    if (temp_pars->temp_parameters.usSerialMode & TWO_STOPS) {
        i = 2;
    }
    else if (temp_pars->temp_parameters.usSerialMode & ONE_HALF_STOPS) {
        i = 1;
    }
    fnDebugMsg((CHAR *)cStops[i]);
    fnDebugMsg(" stop bit(s)\r\n");
}

static int fnGetStops(CHAR *ptr_input, unsigned short *usMode)
{
    int i = 0;
    while (i < NUMBER_OF_STOPS) {
        if (!(uStrcmp((CHAR *)cStops[i], ptr_input))) {
            *usMode &= ~(ONE_HALF_STOPS | TWO_STOPS);                    // default 1 stop bit
            switch (i) {
            case 1:                                                      // 1,5 stop bits
                *usMode |= ONE_HALF_STOPS;
                break;

            case 2:
                *usMode |= TWO_STOPS;                                    // 2 stop bits
                break;
            }
            return 1;                                                    // return found
        }
        i++;
    }
    return 0;
}


#define NUMBER_OF_PARITY 3
static const CHAR cParity[NUMBER_OF_PARITY][5] = {
    {"ODD"},
    {"EVEN"},
    {"NONE"},
};

static void fnShowParity(void)
{
    int i = 2;
    if (temp_pars->temp_parameters.usSerialMode & RS232_EVEN_PARITY) {
        i = 1;
    }
    else if (temp_pars->temp_parameters.usSerialMode & RS232_ODD_PARITY) {
        i = 0;
    }
    fnDebugMsg((CHAR *)cParity[i]);
    fnDebugMsg(" parity\r\n");
}

static int fnGetParity(CHAR *ptr_input, unsigned short *usMode)
{
    int i = 0;
    while (i < NUMBER_OF_PARITY) {
        if (!(uStrcmp(cParity[i], ptr_input))) {
            *usMode &= ~(RS232_EVEN_PARITY | RS232_ODD_PARITY);          // default no parity
            switch (i) {
            case 0:                                                      // odd parity
                *usMode |= RS232_ODD_PARITY;
                break;

            case 1:                                                      // odd parity
                *usMode |= RS232_EVEN_PARITY;
                break;
            }
            return 1;                                                    // return found
        }
        i++;
    }
    return 0;
}

#define NUMBER_OF_FLOW 3
static const CHAR cFlow[NUMBER_OF_FLOW][8] = {
    {"NONE"},
    {"RTS"},
    {"XON"},
};

static void fnShowFlow(void)
{
    int i = 0;
    if (temp_pars->temp_parameters.usSerialMode & RTS_CTS) {
        i = 1;
    }
    else if (temp_pars->temp_parameters.usSerialMode & USE_XON_OFF) {
        i = 2;
    }
    fnDebugMsg("Flow control ");
    fnDebugMsg((CHAR *)cFlow[i]);
    fnDebugMsg("\r\n");
    if (i != 0) {
        fnDebugMsg("High water ");
        fnDebugDec(temp_pars->temp_parameters.ucFlowHigh, 0);
        fnDebugMsg("%\r\n");
        fnDebugMsg("Low water ");
        fnDebugDec(temp_pars->temp_parameters.ucFlowLow, 0);
        fnDebugMsg("%\r\n");
    }
}

static int fnSetFlowControl(CHAR *ptr_input, unsigned short *usMode)
{

    int i = 0;
    while (i < NUMBER_OF_FLOW) {
        if (!(uStrcmp(cFlow[i], ptr_input))) {
            *usMode &= ~(USE_XON_OFF | RTS_CTS);                         // default no flow control
            switch (i) {
            case 1:                                                      // RTC/CTS
                *usMode |= RTS_CTS;
                break;

            case 2:                                                      // XON/XOFF
                *usMode |= USE_XON_OFF;
                break;
            }
            return 1;                                                    // return found
        }
        i++;
    }
    return 0;
}
#endif

#define NO_ENABLE_MATCH 0
#define ENABLE_OPTION   1
#define DISABLE_OPTION  2
static int fnEnable(CHAR *ptrInput)
{
    if (!(uStrcmp("enable", ptrInput)))  return ENABLE_OPTION;
    if (!(uStrcmp("disable", ptrInput))) return DISABLE_OPTION;

    fnDebugMsg("\r\nInvalid switch\r\n");
    return NO_ENABLE_MATCH;
}


static int fnEnableDisableServers(CHAR *ptr_input, unsigned char ucOption)
{
    int iEnable = fnEnable(ptr_input);

    if (iEnable == ENABLE_OPTION) {
        temp_pars->temp_parameters.ucServers |= ucOption;
    }
    else if (iEnable == DISABLE_OPTION) {
        temp_pars->temp_parameters.ucServers &= ~ucOption;
    }
    else {
        return 1;
    }
    return 0;
}

static int fnShowServerEnabled(const CHAR *cText, unsigned char ucOption)
{
    fnDebugMsg((CHAR *)cText);
    fnDebugMsg(" - ");
    if (temp_pars->temp_parameters.ucServers & ucOption) {
        fnDebugMsg("enabled\r\n");
        return 1;
    }
    else {
        fnDebugMsg("disabled\r\n");
        return 0;
    }
}


#if defined SERIAL_INTERFACE && defined DEMO_UART
static void fnShowSerial(void)
{
    fnDebugMsg("\r\nSerial configuration\r\n");
    fnDebugMsg("Baud = ");
    fnShowSpeed();
    fnDebugMsg("\r\n");
    if (temp_pars->temp_parameters.usSerialMode & CHAR_7) {
       fnDebugMsg("7");
    }
    else {
       fnDebugMsg("8");
    }
    fnDebugMsg(" Bits\r\n");
    fnShowParity();
    fnShowStops();
    fnShowFlow();
}
#endif

#if defined SERIAL_INTERFACE && defined DEMO_UART                        // {10}
static void fnDoSerial(unsigned char ucType, CHAR *ptr_input)
{
    int iChangeSerial = 0;

    switch (ucType) {
    case SERIAL_SET_BAUD:
        if (!(fnGetSpeed(ptr_input, &temp_pars->temp_parameters.ucSerialSpeed))) {
           fnDebugMsg("Incorrect serial speed\r\n");
           return;
        }
        iChangeSerial = 1;
        break;

    case SERIAL_SET_DATA_LENGTH:
        if (*ptr_input == '7') {
            temp_pars->temp_parameters.usSerialMode |= CHAR_7;
        }
        else if (*ptr_input == '8') {
           temp_pars->temp_parameters.usSerialMode &= ~CHAR_7;
        }
        else {
           fnDebugMsg("Incorrect data length\r\n");
           return;
        }
        iChangeSerial = 1;
        break;

    case SERIAL_SET_PARITY:
        if (!(fnGetParity(ptr_input, &temp_pars->temp_parameters.usSerialMode))) {
           fnDebugMsg("Incorrect parity\r\n");
           return;
        }
        iChangeSerial = 1;
        break;

    case SERIAL_SET_STOP:
        if (!(fnGetStops(ptr_input, &temp_pars->temp_parameters.usSerialMode))) {
           fnDebugMsg("Incorrect stop bits\r\n");
           return;
        }
        iChangeSerial = 1;
        break;

    case SERIAL_SET_FLOW:
        if (!(fnSetFlowControl(ptr_input, &temp_pars->temp_parameters.usSerialMode))) {
           fnDebugMsg("Incorrect flow control\r\n");
           return;
        }
        iChangeSerial = 1;
        break;

    case SERIAL_SHOW_CONFIG:
        fnShowSerial();
        return;

    case SERIAL_SET_HIGH_WATER:
    case SERIAL_SET_LOW_WATER:
        {
            unsigned char ucValue = (unsigned char)fnDecStrHex(ptr_input);
            if (ucValue > 99) {
                ucValue = 99;
            }
            else if (ucValue == 0) {
                ucValue = 1;
            }
            if (ucType == SERIAL_SET_HIGH_WATER) {
                temp_pars->temp_parameters.ucFlowHigh = ucValue;
            }
            else {
                temp_pars->temp_parameters.ucFlowLow  = ucValue;
            }
            iChangeSerial = 1;
        }
        break;

    default:
        return;
    }

    if (iChangeSerial != 0) {
        fnSetNewSerialMode(MODIFY_CONFIG);                               // if there were changes, modify interface
    }
}
#endif

#ifdef ETH_INTERFACE
    #if defined USE_IP_STATS
static const CHAR cIPStatTypes[TOTAL_OTHER_EVENTS+1][19] = {
    {"Total Rx frames"},
    {"Rx overruns"},
    {"Rx ARP"},
    {"Rx ICMP"},
    {"Rx UDP"},
    {"Rx TCP"},
    {"Rx checksum errors"},
    {"Rx other protocols"},
    {"Foreign Rx ARP"},
    {"Foreign Rx ICMP"},
    {"Foreign Rx UDP"},
    {"Foreign Rx TCP"},
    {"Total Tx frames"},
    {"ARP sent"},
    {"ICMP sent"},
    {"UDP sent"},
    {"TCP sent"},
    {"Other events"}
};
#endif

#define NUMBER_OF_LAN 3
static const CHAR cLAN[NUMBER_OF_LAN][5] = {
    {"10"},
    {"100"},
    {"AUTO"},
};

static void fnShowLAN(NETWORK_PARAMETERS *ptrNetwork)
{
    int i = 2;
    if (ptrNetwork->usNetworkOptions & LAN_10M) {
        i = 0;
    }
    else if (ptrNetwork->usNetworkOptions & LAN_100M) {
        i = 1;
    }
    fnDebugMsg("LAN speed ");
    fnDebugMsg((CHAR *)cLAN[i]);
    fnDebugMsg("\r\n");
    fnShowServerEnabled("DHCP_SERVER", ACTIVE_DHCP);
}

static int fnSetLAN(CHAR *ptr_input, unsigned short *usMode)
{
    int i = 0;
    while (i < NUMBER_OF_LAN) {
        if (!(uStrcmp(cLAN[i], ptr_input))) {
            *usMode &= ~(LAN_10M | LAN_100M);                            // default auto-negotiation
            switch (i) {
            case 0:                                                      // 10M fixed
                *usMode |= LAN_10M;
                break;

            case 1:                                                      // 100M fixed
                *usMode |= LAN_100M;
                break;
            }
            return 1;                                                    // return found
        }
        i++;
    }
    return 0;
}
#endif

#ifdef USE_IP
static void fnDisplayArpEntries(void)
{
    unsigned char ucEntry;
    unsigned char *ptrEntry;

    fnDebugMsg("\r\nARP Table contents:");

    for (ucEntry = 0; ucEntry < ARP_TABLE_ENTRIES; ucEntry++) {
        ptrEntry = fnGetARPentry(ucEntry, GET_IP);
        if (ptrEntry != 0) {                                             // {22}
            fnDebugMsg("\r\nIPV4 address: ");
            fnDisplayIP(ptrEntry);
            fnDebugMsg(" MAC address: ");
            ptrEntry = fnGetARPentry(ucEntry, GET_MAC);
            fnDisplayMAC(ptrEntry);
        }
    }
    fnDebugMsg("\r\nEnd\r\n");
}

static void fnIpConfig(NETWORK_PARAMETERS *ptrNetwork)
{
    fnDebugMsg("\r\nIPV4 address = ");
    fnDisplayIP(&ptrNetwork->ucOurIP[0]);
#ifdef USE_IPV6                                                          // {21}
    fnDebugMsg("\r\nIPV6 address = ");
    fnDisplayIPV6(&ptrNetwork->ucOurIPV6[0]);
#endif
    fnDebugMsg("\r\nMAC address = ");
    fnDisplayMAC(&ptrNetwork->ucOurMAC[0]);
    fnDebugMsg("\r\nSubnet mask = ");
    fnDisplayIP(&ptrNetwork->ucNetMask[0]);
    fnDebugMsg("\r\nDefault gateway = ");
    fnDisplayIP(&ptrNetwork->ucDefGW[0]);
    fnDebugMsg("\r\n");
}
#endif

#ifdef USE_IPV6
static void fnDisplayNeighborEntries(void)
{
    unsigned char ucEntry;
    unsigned char *ptrEntry;

    fnDebugMsg("\r\nIPV6 neighbor Table contents:");

    for (ucEntry = 0; ucEntry < NEIGHBOR_TABLE_ENTRIES; ucEntry++) {
        ptrEntry = fnGetNeighborEntry(ucEntry, GET_IP);
        if (ptrEntry != 0) {
            fnDebugMsg("\r\nIPV6 address: ");
            fnDisplayIPV6(ptrEntry);
            fnDebugMsg(" MAC address: ");
            ptrEntry = fnGetNeighborEntry(ucEntry, GET_MAC);
            fnDisplayMAC(ptrEntry);
        }
    }
    fnDebugMsg("\r\nEnd\r\n");
}
#endif

static void fnDoIP(unsigned char ucType, CHAR *ptr_input)
{
    switch (ucType) {
#if defined USE_ICMP && defined ICMP_SEND_PING
    case PING_TEST:
        fnStrIP(ptr_input, ucTempIP);                                    // ping test address
        fnPingTest(-2);
        return;
#endif
#ifdef USE_IP
    case DO_SET_IP_ADDRESS:
        fnStrIP(ptr_input, &temp_pars->temp_network.ucOurIP[0]);         // set device's IP address to the temporary set
        break;
#endif
#ifdef USE_IPV6
    case DO_SET_IPV6_ADDRESS:
    case PING_TEST_V6:
        {
            unsigned char ucTestIPV6[IPV6_LENGTH];                       // extract the address locally in case it has syntax errors
            if (fnStrIPV6(ptr_input, ucTestIPV6) != 0) {
                if (ucType == PING_TEST_V6) {
    #ifdef ICMP_SEND_PING
                    uMemcpy(&ucPingTestV6, ucTestIPV6, IPV6_LENGTH);     // save the IP address to be tested
                    fnPingV6Test(-1);                                    // start an IPV6 ping test
    #endif
                }
                else {
                    uMemcpy(&temp_pars->temp_network.ucOurIPV6[0], ucTestIPV6, IPV6_LENGTH); // set the new temporary IPV6 address
                }
            }
            else {
                fnDebugMsg("Invalid IPV6 entry!\r\n");
            }
        }
        break; 
#endif
#ifdef ETH_INTERFACE
    case DO_SET_MAC:                                                     // set device's MAC address
        if (((uMemcmp(&network.ucOurMAC[0], cucNullMACIP, MAC_LENGTH))) || (!(fnSetMAC(ptr_input, &temp_pars->temp_network.ucOurMAC[0])))) {                                       // interpret MAC address input and save this
            fnDebugMsg("MAC may not be modified!!\r\n");                 // we are only allowed to save the new address once
            return;
        }
    #ifdef USE_PARAMETER_BLOCK                                           // {5}
        else {
            fnDoFlash(DO_SAVE_PARS, 0);                                  // save the new MAC address
            fnResetBoard();                                              // reset so that it strats with the new values
        }
    #endif
        break;
#endif
#ifdef USE_IP
    case DO_SET_IP_SUBNET_MASK:                                          // set device's subnet mask
        fnStrIP(ptr_input, &temp_pars->temp_network.ucNetMask[0]);
        break;

    case DO_SET_DEFAULT_GATEWAY:                                         // set device's default gateway
        fnStrIP(ptr_input, &temp_pars->temp_network.ucDefGW[0]);
        break;

    case DO_SHOW_IP_CONFIG_TEMP:                                         // show the present working IP configuration
        fnIpConfig(&network);
        fnShowLAN(&network);
        break;

    case DO_SHOW_IP_CONFIG:                                              // show the temporary working IP configuration which may be saved
        fnIpConfig(&temp_pars->temp_network);
        fnShowLAN(&temp_pars->temp_network);
        break;
#endif
#ifdef USE_IP_STATS
    case DO_RESET_ETHERNET_STATS:                                        // {13}
        fnDeleteEthernetStats();
        break;

    case DO_SHOW_ETHERNET_STATS:                                         // show Ethernet statistic counters
        fnDebugMsg("\r\nEthernet Statistics\r\n");
        for (ucType = 0; ucType <= TOTAL_OTHER_EVENTS; ucType++) {
            fnDebugMsg("\r\n");
            fnDebugMsg((CHAR *)cIPStatTypes[ucType]);
            fnDebugMsg(" = ");
            fnDebugDec(fnGetEthernetStats(ucType), 0);
        }
        fnDebugMsg("\r\n");
        return;
#endif
#ifdef ETH_INTERFACE
    case DO_SET_LAN_SPEED:
        if (!(fnSetLAN(ptr_input, &temp_pars->temp_network.usNetworkOptions /* &parameters->usSerialMode*/))) {
           fnDebugMsg("Incorrect LAN speed\r\n");
           return;
        }
        break;
#endif
#ifdef USE_IP
    case SHOW_ARP:
        fnDisplayArpEntries();                                           // display present ARP table contents
        return;

    case DELETE_ARP:                                                     // delete the ARP table contents
        fnDeleteArp();
        break;
#endif
#ifdef USE_IPV6
    case SHOW_NN:
        fnDisplayNeighborEntries();                                      // display present IPV6 neighbor table contents
        return;

    case DELETE_NN:                                                      // delete the IPV6 neighbor table contents
        fnDeleteNeighbors();
        break;
#endif
    }
}


static void fnDoAdmin(unsigned char ucType, CHAR *ptrInput)
{
    switch (ucType) {

    case DO_SHOW_CONFIG:
        fnDisplaySerialNumber();
        fnDisplaySoftwareVersion();
        fnDisplayDeviceName();
        break;

    case DO_SET_DEVICE_NAME:
        uMemcpy(temp_pars->temp_parameters.cDeviceIDName, ptrInput, (sizeof(temp_pars->temp_parameters.cDeviceIDName)-1));
        break;

#ifdef TEST_CLIENT_SERVER
    case DO_CONNECT:                                                     // connect to remote server and send test frames
        fnTestTCPServer(CONNECT_REMOTE);                                 // connect to a remote client
        break;

    case DO_START:
        fnTestTCPServer(TCP_SERVER_TX);                                  // start server test
        break;
#endif
    }
}

static void fnShowSec(void)
{
    fnDebugMsg("\r\nSecurity settings\r\n");
    fnShowServerEnabled("Terminal menu login", ACTIVE_LOGIN);
    fnShowServerEnabled("FTP server", ACTIVE_FTP_SERVER);
    if (fnShowServerEnabled("WEB server", ACTIVE_WEB_SERVER)) {
        fnShowServerEnabled("WEB server authentication", AUTHENTICATE_WEB_SERVER);
    }
    if (fnShowServerEnabled("TELNET server", ACTIVE_TELNET_SERVER)) {
        fnDebugMsg("   Telnet port number = ");
        fnDebugDec(temp_pars->temp_parameters.usTelnetPort, WITH_CR_LF);
    }
}


static void fnDoServer(unsigned char ucType, CHAR *ptrInput)
{
    switch (ucType) {
    case DO_ENABLE_TELNET:
        if (fnEnableDisableServers(ptrInput, ACTIVE_TELNET_SERVER)) {
            return;
        }
        fnConfigureTelnetServer();
        break;

    case DO_ENABLE_DHCP:
        if (fnEnableDisableServers(ptrInput, ACTIVE_DHCP)) {
            return;
        }
        break;

    case DO_ENABLE_FTP:
        if (fnEnableDisableServers(ptrInput, ACTIVE_FTP_SERVER)) {
            return;
        }
        fnConfigureFtpServer(FTP_TIMEOUT);                               // activate / deactivate server {3}
        break;

    case DO_ENABLE_LOGIN:
        if (fnEnableDisableServers(ptrInput, ACTIVE_LOGIN)) {
            return;
        }
        break;

    case DO_ENABLE_WEB_SERVER:
        if (fnEnableDisableServers(ptrInput, ACTIVE_WEB_SERVER)) {
            return;
        }
 #ifdef USE_HTTP
        fnConfigureAndStartWebServer();
 #endif
        break;

    case DO_HTTP_AUTH:
        if (fnEnableDisableServers(ptrInput, AUTHENTICATE_WEB_SERVER)) {
            return;
        }
 #ifdef USE_HTTP
        fnConfigureAndStartWebServer();
 #endif
        break;

    case SHOW_SECURITY:
        fnShowSec();
        break;

    default:
        return;
    }
}

static void fnDoTelnet(unsigned char ucType, CHAR *ptrInput)
{
    switch (ucType) {

      case DO_SET_TELNET_PORT:
          temp_pars->temp_parameters.usTelnetPort = (unsigned short)fnDecStrHex(ptrInput);
          break;

      case DO_TELNET_QUIT:
          if (usTelnet_state == ES_NETWORK_COMMAND_MODE){
              fnDebugMsg("\n\rBye Bye\n\r");
              usTelnet_state = ES_TERMINATE_CONNECTION;                  // now quit the Telnet session
          }
          else {
#if defined USE_MAINTENANCE || defined USB_INTERFACE                     // {8}
              usUSB_state = 0;
#endif
              fnGotoNextState(ES_NO_CONNECTION);                         // log out of serial session and kill any monitoring timers
          }
          return;
    }
}




// This routine is used when we know that we have to jump some white space
//
static int fnJumpWhiteSpace(CHAR **ptrptrInput)
{
    CHAR *ptrInput = *ptrptrInput;

    if (*ptrInput != ' ') {
        while (*ptrInput != 0) {
            if (*ptrInput != ' ') {
                ptrInput++;
            }
            else {
                break;
            }
        }
    }

    while (*ptrInput != 0) {
        if (*ptrInput == ' ') {
            ptrInput++;
        }
        else {
            break;
        }
    }
    *ptrptrInput = ptrInput;
    return (*ptrInput == 0);
}


static int fnSetPort(unsigned char ucType, CHAR *ptrInput)               // modify the port type
{
    unsigned char ucBit;

    switch (ucType) {
    case DO_DDR:
        ucBit = (unsigned char )fnHexStrHex(ptrInput);                   // extract the bit to be changed (1..4)
        if (fnJumpWhiteSpace(&ptrInput)) return -1;
        return (fnConfigPort((CHAR)(ucBit + '0' - 1), (*ptrInput)));
    }
    return 0;
}


static void fnDoHardware(unsigned char ucType, CHAR *ptrInput)
{
    switch (ucType) {
    case DO_RESET:                                                       // hardware - reset
        fnResetBoard();
        break;

    case DO_DISPLAY_MEMORY_USE:                                          // memory use display
        fnDisplaySystemMemory();
        break;

    case DO_DISPLAY_UP_TIME:
        {
            CHAR cUpTime[20];
            *(fnUpTime(cUpTime)) = 0;
            fnDebugMsg("\r\nThe device has been operating for ");
            fnDebugMsg(cUpTime);
        }
        return;

//    case DO_GET_USER_OUTPUT:                                           // display whether default or user - removed {2}
//        if ((*ptrInput < '0') || (*ptrInput > '3')) {
//            fnDebugMsg("Bad port number\r\n");
//            break;
//        }
//        fnDebugMsg("\r\nPort use: ");
//        if (!fnPortInputConfig(*ptrInput)) {
//            fnDebugMsg("USER\r\n");
//        }
//        else {
//            fnDebugMsg("DEFAULT\r\n");
//        }
//        break;

      case DO_GET_DDR:                                                   // get present ddr setting {2}
          if ((*ptrInput < '1') || (*ptrInput > '4')) {
              fnDebugMsg("Bad port number\r\n");
              break;
          }
          fnDebugMsg("\r\nPort direction: ");
          if (!fnPortInputConfig((*ptrInput))) {
              fnDebugMsg("OUTPUT\r\n");
          }
          else {
              fnDebugMsg("INPUT\r\n");
          }
          break;

      case DO_OUTPUT:                                                    // set output port value {2}
          if ((*ptrInput < '1') || (*ptrInput > '4')) {
              fnDebugMsg("Invalid port\r\n");
              return;
          }
          if (fnPortInputConfig(*ptrInput)) {
              fnDebugMsg("Port is input\r\n");
              return;
          }
          else {
              unsigned char ucPresentPort = 0;
              unsigned char ucBit = 0x01;
              CHAR cPortBit = '1';
              while (cPortBit < '5') {
                  if (fnPortState(cPortBit++)) {
                      ucPresentPort |= ucBit;
                  }
                  ucBit <<= 1;
              }
              ucBit = (unsigned char)fnHexStrHex(ptrInput);
              ucBit = (1<<(ucBit-1));
              if (fnHexStrHex(++ptrInput) == 0x01) {
                  ucPresentPort |= ucBit;
              }
              else {
                  ucPresentPort &= ~ucBit;
              }
              fnSetPortOut(ucPresentPort, 0);                            // set new port state
          }
          break;

//    case DO_SET_USER_OUTPUT:                                           // set either default or user - removed {2}
//        if (fnConfigPort(*ptrInput, *(ptrInput+2)) < 0) {
//            fnDebugMsg("Error in input\r\n");
//            return;
//        }
//        break;

      case DO_SAVE_PORT:
          fnSavePorts();
          break;

      case DO_INPUT:                                                     // read an input bit {2}
            if ((*ptrInput < '1') || (*ptrInput > '4')) {
                fnDebugMsg("Invalid port\r\n");
                return;
            }
            fnDebugMsg("Port state: ");
            if (fnPortState(*ptrInput)) {
              fnDebugMsg("1\r\n");
            }
            else {
              fnDebugMsg("0\r\n");
            }
            break;

      case DO_DDR:                                                       // set port direction (and use)
          if (fnSetPort(ucType, ptrInput)) {                             // modify the port
              fnDebugMsg("\r\n??\r\n#");                                 // signal input not recognised and send prompt in command mode
              return;
          }
          break;
    }
}

#ifdef USB_INTERFACE                                                     // {8}
static void fnDoUSB(unsigned char ucType, CHAR *ptrInput)                // USB group
{
    switch (ucType) {
    case DO_USB_RS232_MODE:                                              // set RS232<->USB link. Exit is via USB disconnect
    #ifdef SERIAL_INTERFACE
        fnDebugMsg("Connecting USB with RS232 - disconnect USB to quit...\r\n");
        usUSB_state |= ES_USB_RS232_MODE;
    #endif
        break;
    case DO_USB_DOWNLOAD:                                                // start a download via USB
        if (usUSB_state != ES_USB_COMMAND_MODE) {
            fnDebugMsg("Please use USB connection!!\r\n");
        }
        else {
    #ifdef M52210DEMO
            fnDebugMsg("Download not supported on M52210DEMO!!\r\n");
    #else
            fnDebugMsg("Please start download:");
            usUSB_state |= ES_USB_DOWNLOAD_MODE;
    #endif
        }
        break;
    }
}
#endif

#ifdef SUPPORT_OLED
extern void fnSaveGrayScales(void);
extern unsigned char ucGrayScales[15 + 1];
static void fnDoOLED(unsigned char ucType, CHAR *ptrInput)               // OLED group
{
    switch (ucType) {
    case DO_OLED_GET_GRAY_SCALES:                                        // get gray levels
        {
            int x;
            fnDebugMsg("Present settings:\r\n");
            for (x = 1; x < 16; x++) {
                fnDebugMsg("GS");
                fnDebugDec(x, 0);
                fnDebugMsg(" = ");
                fnDebugDec(ucGrayScales[x], WITH_CR_LF);
            }
        }
        break;
    case DO_OLED_SET_GRAY_SCALE:                                         // set gray level value
        {
            unsigned char ucGrayScaleValue;
            unsigned char ucMin = 0;
            unsigned char ucMax = 64;
            unsigned char ucGrayScaleIndex = (unsigned char)fnDecStrHex(ptrInput);
            if ((ucGrayScaleIndex < 1) || (ucGrayScaleIndex > 15) || (fnJumpWhiteSpace(&ptrInput))) {
                fnDebugMsg("Invalid GS!\r\n");
                return;
            }
            ucGrayScaleValue = (unsigned char)fnDecStrHex(ptrInput);
            if (ucGrayScaleIndex != 1) {
                ucMin = ucGrayScales[ucGrayScaleIndex - 1];
            }
            if (ucGrayScaleIndex != 15) {
                ucMax = ucGrayScales[ucGrayScaleIndex + 1];
            }
            if ((ucGrayScaleValue <= ucMin) || (ucGrayScaleValue >= ucMax)) {
                fnDebugMsg("Invalid value!\r\n");
            }
            else {
                ucGrayScales[ucGrayScaleIndex] = ucGrayScaleValue;
                fnSaveGrayScales();
                fnDebugMsg("OK\r\n");
            }
        }
        break;
    }
}
#endif

#if defined IIC_INTERFACE && defined TEST_IIC_INTERFACE
static void fnDoI2C(unsigned char ucType, CHAR *ptrInput)                // I2C group
{
    switch (ucType) {
    case DO_I2C_WRITE:                                                   // write a value to a specified I2C address
        {
            unsigned char ucTest[3]; // = {IIC_WRITE_ADDRESS, 0, 0};
            ucTest[0] = IIC_WRITE_ADDRESS;
            ucTest[1] = (unsigned char)fnHexStrHex(ptrInput);            // the address
            if (fnJumpWhiteSpace(&ptrInput) != 0) {
                fnDebugMsg("Data missing\r\n");
                return;
            }
            ucTest[2] = (unsigned char)fnHexStrHex(ptrInput);            // the data
            fnWrite(IICPortID, (unsigned char *)ucTest, sizeof(ucTest));
            fnDebugMsg("Written value ");
            fnDebugHex(ucTest[2], (WITH_LEADIN | sizeof(ucTest[2])));
            fnDebugMsg(" to address ");
            fnDebugHex(ucTest[1], (WITH_LEADIN | WITH_CR_LF | sizeof(ucTest[1])));
        }
        break;

    case DO_I2C_READ_PLUS_WRITE:
    case DO_I2C_READ:                                                    // read values from an I2C address
        {
            unsigned char ucSetAddress[2];// = {IIC_WRITE_ADDRESS, 0};
            unsigned char ucTest[3];// = {1, IIC_READ_ADDRESS, OWN_TASK};
            ucSetAddress[0] = IIC_WRITE_ADDRESS;
            ucSetAddress[1] = (unsigned char)fnHexStrHex(ptrInput);      // the address
            fnWrite(IICPortID, (unsigned char *)ucSetAddress, sizeof(ucSetAddress));
            if (fnJumpWhiteSpace(&ptrInput) == 0) {
                ucTest[0] = (unsigned char)fnDecStrHex(ptrInput);        // the length               
            }
            else {
                ucTest[0] = 1;
            }
            ucTest[1] = IIC_READ_ADDRESS;
            ucTest[2] = OWN_TASK;
            fnRead(IICPortID, (unsigned char *)ucTest, 0);               // start the read
            if (DO_I2C_READ_PLUS_WRITE == ucType) {
                fnWrite(IICPortID, (unsigned char *)ucSetAddress, sizeof(ucSetAddress));
                fnDebugMsg("Reading (+ wr) ");
            }
            else {
                fnDebugMsg("Reading ");
            }
            fnDebugDec(ucTest[0], sizeof(ucTest[1]));
            fnDebugMsg(" bytes from address ");
            fnDebugHex(ucSetAddress[1], (WITH_LEADIN | WITH_CR_LF | sizeof(ucSetAddress[1])));
        }
        break;

    case DO_I2C_READ_NO_ADDRESS:
        {
            unsigned char ucTest[3];// = {1, IIC_READ_ADDRESS, OWN_TASK};
            ucTest[0] = (unsigned char)fnDecStrHex(ptrInput);            // the length               
            ucTest[1] = IIC_READ_ADDRESS;
            ucTest[2] = OWN_TASK;
            fnRead(IICPortID, (unsigned char *)ucTest, 0);               // start the read
            fnDebugMsg("Reading ");
            fnDebugDec(ucTest[0], sizeof(ucTest[1]));
            fnDebugMsg(" bytes from present address\n\r");
        }
        break;
    }
}
#endif

#ifdef SDCARD_SUPPORT                                                    // {17}

// Handle SD-card disk commands
//
static int fnDoDisk(unsigned char ucType, CHAR *ptrInput)
{
    static UTLISTDIRECTORY utListDirectory;                              // list directory object for a single user
    if ((DO_FORMAT != ucType) && (DO_DISPLAY_SECTOR != ucType) && (DO_INFO != ucType)) {
        if (ptr_utDirectory == 0) {
            ptr_utDirectory = utAllocateDirectory(DISK_D, UT_PATH_LENGTH); // allocate a directory for use by this module associated with D: and reserve its path name string length
        }
        if (!(ptr_utDirectory->usDirectoryFlags & UTDIR_VALID)) {        // directory not valid
            if (utOpenDirectory(0, ptr_utDirectory) != UTFAT_SUCCESS) {  // open the root directory
                fnDebugMsg("No SD-Card ready\r\n");
                return 0;
            }
            utListDirectory.ptr_utDirObject = ptr_utDirectory;           // the list directory is always referenced to the main directory object
        }
    }
    switch (ucType) {
    case DO_DIR:                                                         // display present directory
        if ((iFATstalled == 0) && (utLocateDirectory(ptrInput, &utListDirectory) < UTFAT_SUCCESS)) { // open a list referenced to the main directory
            fnDebugMsg("Invalid directory\r\n");
        }
        else {                                                           // list the directory content
            static FILE_LISTING fileList = {0};
            UTDISK *ptrDiskInfo;
            int iResult;
            CHAR cBuffer[MAX_UTFAT_FILE_NAME + DOS_STYLE_LIST_ENTRY_LENGTH]; // temporary string buffer for listing
            fileList.usMaxItems = 1;
            fileList.ptrBuffer = cBuffer;
            fileList.usBufferLength = sizeof(cBuffer);
            fileList.ucStyle = DOS_TYPE_LISTING;
            if (iFATstalled == 0) {
                fnDebugMsg("Directory ");
                fnDebugMsg(ptr_utDirectory->ptrDirectoryPath);
                if (ptr_utDirectory->usRelativePathLocation > 3) {       // if the relative path is not the root directory
                    fnDebugMsg("\\");
                }
                if ((ptrInput != 0) &&  (*ptrInput != 0)) {
                    fnDebugMsg(ptrInput);
                }
                fnDebugMsg("\r\n\n");
            }
            iFATstalled = 0;
            while (1) {
                if (!fnWrite(DebugHandle, 0, sizeof(cBuffer))) {         // check whether there is enough space in the output buffer to accept the next entry
                    fnDriver(DebugHandle, MODIFY_WAKEUP, (MODIFY_TX | OWN_TASK)); // we want to be woken when the queue is free again
                    iFATstalled = STALL_DIR_LISTING;
                    return 1;
                }
                iResult = utListDir(&utListDirectory, &fileList);
                if (fileList.usStringLength != 0) {
                    fnWrite(DebugHandle, (unsigned char *)cBuffer, (QUEUE_TRANSFER)fileList.usStringLength);
                }
                if (iResult == UTFAT_NO_MORE_LISTING_ITEMS_FOUND) {
                    break;                                               // end of the directory content
                }
            }
            fnDebugDec(fileList.usFileCount, 0);
            fnDebugMsg(" files with ");
            fnDebugDec(fileList.ulFileSizes, 0);
            fnDebugMsg(" bytes\r\n");
            fnDebugDec(fileList.usDirectoryCount, 0);
            fnDebugMsg(" directories, ");
            uMemset(&fileList, 0, sizeof(fileList));                     // reset for next use
            ptrDiskInfo = fnGetDiskInfo(DISK_D);
            if ((ptrDiskInfo->ucDiskFlags & FSINFO_VALID) && (ptrDiskInfo->utFileInfo.ulFreeClusterCount != 0xffffffff)) {
                fnDebugDec((ptrDiskInfo->utFileInfo.ulFreeClusterCount * ptrDiskInfo->utFAT.ucSectorsPerCluster * ptrDiskInfo->utFAT.usBytesPerSector), 0);
                fnDebugMsg(" bytes free\r\n");
            }
            else {
                iFATstalled = STALL_COUNTING_CLUSTERS;
                utFreeClusters(DISK_D, OWN_TASK);                        // start count of free clusters - the result will be displayed on completion
                return 1;
            }
        }
        break;

    case DO_WRITE_FILE:                                                  // write extra content to file
    case DO_PRINT_FILE:                                                  // print the content of a file (ASCII)
        {
            static UTFILE utFile;                                        // local file object
            unsigned char ucTemp[256];                                   // temp buffer to retrieve a block of data from the file
            utFile.ptr_utDirObject = ptr_utDirectory;
            utFile.ownerTask = OWN_TASK;
            if ((iFATstalled == STALL_PRINTING_FILE) || (utOpenFile(ptrInput, &utFile, (UTFAT_OPEN_FOR_READ | UTFAT_OPEN_FOR_WRITE | UTFAT_MANAGED_MODE)) == UTFAT_PATH_IS_FILE)) { // open a file referenced to the directory object
                int iLoop;
                if (DO_WRITE_FILE == ucType) {                           // add input to file
                    static unsigned char ucContent = 0x55;
                    uMemset(ucTemp, ucContent, sizeof(ucTemp));          // fill with a pattern
                    ucContent++;                                         // change for next write
                    utSeek(&utFile, 0, UTFAT_SEEK_END);                  // move to the end of the file
                    if (utWriteFile(&utFile, ucTemp, sizeof(ucTemp)) != UTFAT_SUCCESS) {
                        fnDebugMsg("write failed");
                    }
                    else {
                        fnDebugMsg("New file length = ");
                        fnDebugDec(utFile.ulFileSize, 0);
                    }
                    fnDebugMsg("\r\n");
                    utCloseFile(&utFile);
                    break;
                }
                do {
                    if (!fnWrite(DebugHandle, 0, sizeof(ucTemp))) {      // check whether there is enough space in the output buffer to accept the next block
                        fnDriver(DebugHandle, MODIFY_WAKEUP, (MODIFY_TX | OWN_TASK)); // we want to be woken when the queue is free again
                        iFATstalled = STALL_PRINTING_FILE;
                        return 1;
                    }
                    iFATstalled = 0;
                    if (utReadFile(&utFile, ucTemp, sizeof(ucTemp)) != UTFAT_SUCCESS) {
                        fnDebugMsg("\r\nREAD ERROR occured");
                        break;
                    }
                    for (iLoop = 0; iLoop < utFile.usLastReadWriteLength; iLoop++) {
                        if ((ucTemp[iLoop] < 0x0a) || (ucTemp[iLoop] > 'z')) {
                            ucTemp[iLoop] = '.';                         // non-printable characters replaced
                        }
                    }
                    fnWrite(DebugHandle, ucTemp, utFile.usLastReadWriteLength); // print the read data buffer
                    if (utFile.usLastReadWriteLength < sizeof(ucTemp)) { // complete file content has been read
                        break;
                    }
                } while (1);
                fnDebugMsg("\r\n");
                utCloseFile(&utFile);
            }
            else {
                fnDebugMsg("File not found\r\n");
            }
        }
        break;

//  case DO_ROOT:                                                        // {19}
//      ptr_utDirectory->ucDirectoryFlags &= ~UTDIR_VALID;               // reset validity flag
//      if (utOpenDirectory(0, ptr_utDirectory) != UTFAT_PATH_IS_ROOT) { // open the root directory
//          fnDebugMsg("No SD-Card ready\r\n");
//          return 0;
//      }
//      break;
    case DO_DISPLAY_SECTOR:
        {
            int i, j;
            unsigned long ulSectorNumber = fnHexStrHex(ptrInput);        // the sector number
            unsigned long ulBuffer[512/sizeof(unsigned long)];
            fnDebugMsg("Reading sector ");
            fnDebugHex(ulSectorNumber, (WITH_LEADIN | sizeof(ulSectorNumber)));
            if (fnReadSector(DISK_D, (unsigned char *)ulBuffer, ulSectorNumber) != 0) {
                fnDebugMsg(" FAILED!!\r\n");
                break;
            }
            fnDebugMsg("\r\n");
            for (i = 0; i < 8; i++) {
                for (j = 0; j < 16; j++) {
                    fnDebugHex(ulBuffer[(i * 16) + j], (WITH_LEADIN | WITH_SPACE | sizeof(ulBuffer[0])));
                }
                fnDebugMsg("\r\n");
            }
        }
        break;
#ifdef UTFAT_FORMATTING
    case DO_REFORMAT:
    case DO_FORMAT:
        {
            fnDebugMsg("Formatting ");
            if (DO_REFORMAT == ucType) {
                utReFormat(DISK_D, ptrInput);
            }
            else if (utFormat(DISK_D, ptrInput) != UTFAT_SUCCESS) {
                fnDebugMsg("not possible\r\n");
                break;
            }
            fnDebugMsg("in progress - please wait...\r\n");
        }
        break;
#endif
    case DO_NEW_DIR:
        if (utMakeDirectory(ptrInput, ptr_utDirectory) != UTFAT_SUCCESS) {
            fnDebugMsg("Make dir failed\r\n");
        }
        break;
    case DO_NEW_FILE:
        {
            UTFILE utFile;                                               // temporary file object for creation
            utFile.ptr_utDirObject = ptr_utDirectory;
            if (utOpenFile(ptrInput, &utFile, (UTFAT_OPEN_FOR_WRITE | UTFAT_CREATE | UTFAT_TRUNCATE)) != UTFAT_PATH_IS_FILE) {// open and truncate file or create it if not existing
                fnDebugMsg("Create file failed\r\n");
            }
        }
        break;
    case DO_RENAME:
        {
            UTFILE utFile;
            CHAR *ptrNewName = ptrInput;
            while (*ptrNewName != ' ') {
                if (*ptrNewName++ == 0) {
                    fnDebugMsg("New name missing\r\n");
                }
            }
            *ptrNewName++ = 0;                                           // intermediate terminator
            utFile.ptr_utDirObject = ptr_utDirectory;
            if (utOpenFile(ptrInput, &utFile, UTFAT_OPEN_FOR_RENAME) < UTFAT_SUCCESS) { // {20}
                fnDebugMsg("Not found\r\n");
            }
            else {
                if (utRenameFile(ptrNewName, &utFile) != UTFAT_SUCCESS) {
                    fnDebugMsg("Rename failed\r\n");
                }
            }
        }
        break;
    case DO_DELETE:
        {
            int iResult = utDeleteFile(ptrInput, ptr_utDirectory);
            if (iResult == UTFAT_DIR_NOT_EMPTY) {
                fnDebugMsg("Dir NOT empty\r\n");
            }
            else if (iResult != UTFAT_SUCCESS) {
                fnDebugMsg("Delete failed\r\n");
            }
        }
        break;
    case DO_CHANGE_DIR:
        if (utChangeDirectory(ptrInput, ptr_utDirectory) != UTFAT_SUCCESS) {// change the directory location
            fnDebugMsg("Invalid path\r\n");
        }
        break;
    case DO_INFO:
        {
            UTDISK *ptrDiskInfo = fnGetDiskInfo(DISK_D);
            fnDebugMsg("\r\nSD-card ");
            if (ptrDiskInfo->ucDiskFlags == 0) {                         // {24} no card detected
                fnDebugMsg("not detected\r\n");
            }
            else {
                int i = 0;
                if (!(ptrDiskInfo->ucDiskFlags & DISK_FORMATTED)) {
                    fnDebugMsg("not formatted");
                }
                else {
                    CHAR cVolumeLabel[12];
                    while (ptrDiskInfo->cVolumeLabel[i] != ' ') {
                        cVolumeLabel[i] = ptrDiskInfo->cVolumeLabel[i];                        
                        if (++i >= 11) {
                            break;
                        }
                    }
                    cVolumeLabel[i] = 0;
                    fnDebugMsg(cVolumeLabel);
                }
                fnDebugMsg(" (");
                fnDebugDec((ptrDiskInfo->ulSD_sectors * 512), 0);
                fnDebugMsg(" bytes)");
                if (ptrDiskInfo->ucDiskFlags & HIGH_CAPACITY_SD_CARD) {
                    fnDebugMsg(" SDHC");
                }
                if (ptrDiskInfo->ucDiskFlags & WRITE_PROTECTED_SD_CARD) {
                    fnDebugMsg(" write-protected");
                }
                fnDebugMsg("\r\n");
                if (ptrDiskInfo->ucDiskFlags & DISK_FORMATTED) {
                    fnDebugMsg("Bytes per sector:");
                    fnDebugDec(ptrDiskInfo->utFAT.usBytesPerSector, (WITH_SPACE | WITH_CR_LF));
                    fnDebugMsg("Cluster size:");
                    fnDebugDec((ptrDiskInfo->utFAT.ucSectorsPerCluster * ptrDiskInfo->utFAT.usBytesPerSector), (WITH_SPACE | WITH_CR_LF));
                    fnDebugMsg("Directory base:");
                    fnDebugHex(ptrDiskInfo->ulDirectoryBase, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->ulDirectoryBase)));
                    fnDebugMsg("FAT start:");
                    fnDebugHex(ptrDiskInfo->utFAT.ulFAT_start, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->utFAT.ulFAT_start)));
                    fnDebugMsg("FAT size:");
                    fnDebugHex(ptrDiskInfo->utFAT.ulFatSize, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->utFAT.ulFatSize)));
                    fnDebugMsg("Number of FATs:");
                    fnDebugDec(ptrDiskInfo->utFAT.ucNumberOfFATs, (WITH_SPACE | WITH_CR_LF));
                    fnDebugMsg("LBA:");
                    fnDebugHex(ptrDiskInfo->ulLogicalBaseAddress, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->ulLogicalBaseAddress)));
                    fnDebugMsg("Total clusters:");
                    fnDebugHex(ptrDiskInfo->utFAT.ulClusterCount, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->utFAT.ulClusterCount)));
                    fnDebugMsg("Info sect:");
                    fnDebugHex(ptrDiskInfo->utFileInfo.ulInfoSector, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->utFileInfo.ulInfoSector)));
                    fnDebugMsg("Free clusters:");
                    fnDebugHex(ptrDiskInfo->utFileInfo.ulFreeClusterCount, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->utFileInfo.ulFreeClusterCount)));
                    fnDebugMsg("Next free:");
                    fnDebugHex(ptrDiskInfo->utFileInfo.ulNextFreeCluster, (WITH_SPACE | WITH_CR_LF | WITH_LEADIN | sizeof(ptrDiskInfo->utFileInfo.ulNextFreeCluster)));
                }
                fnDebugMsg("CSD:");
                for (i = 0; i < sizeof(ptrDiskInfo->utFileInfo.ucCardSpecificData); i++) {
                    fnDebugHex(ptrDiskInfo->utFileInfo.ucCardSpecificData[i], (WITH_SPACE | WITH_LEADIN | sizeof(ptrDiskInfo->utFileInfo.ucCardSpecificData[0])));
                }
                fnDebugMsg("\r\n");
            }
        }
        break;
    }
    return 0;
}
#endif

static int fnDoCommand(unsigned char ucFunction, unsigned char ucType, CHAR *ptrInput)
{
    switch (ucFunction) {                                                // decide which function type to call
    case DO_HELP:                                                        // help group
        if (DO_MAIN_HELP == ucType) {
            return (fnDisplayHelp(0));                                   // show present menu
        }
        else if (DO_MENU_HELP_LAN == ucType) {
            ucMenu = MENU_HELP_LAN;                                      // set LAN menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_MENU_HELP_SERIAL == ucType) {
            ucMenu = MENU_HELP_SERIAL;                                   // set SERIAL menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_MENU_HELP_IO == ucType) {
            ucMenu = MENU_HELP_IO;                                       // set I/O menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_MENU_HELP_ADMIN == ucType) {
            ucMenu = MENU_HELP_ADMIN;                                    // set ADMIN menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_MENU_HELP_STATS == ucType) {
            ucMenu = MENU_HELP_STATS;                                    // set statistics menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_MENU_HELP_USB == ucType) {
            ucMenu = MENU_HELP_USB;                                      // set USB menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_MENU_HELP_I2C == ucType) {
            ucMenu = MENU_HELP_I2C;                                      // set I2C menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_MENU_HELP_DISK == ucType) {                          // {17}
            ucMenu = MENU_HELP_DISK;                                     // set SD-card disk menu
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        else if (DO_HELP_UP == ucType) {
            ucMenu = 0;                                                  // set top menu level
            return (fnDisplayHelp(0));                                   // large menu may require special handling
        }
        break;

    case DO_HARDWARE:                                                    // hardware group
        fnDoHardware(ucType, ptrInput);
        break;
#ifdef USB_INTERFACE
    case DO_USB:
        fnDoUSB(ucType, ptrInput);                                       // USB group
        break;
#endif

#if defined IIC_INTERFACE && defined TEST_IIC_INTERFACE
    case DO_I2C:
        fnDoI2C(ucType, ptrInput);                                       // I2C group
        break;
#endif

#ifdef SDCARD_SUPPORT                                                    // {17}
    case DO_DISK:
        return (fnDoDisk(ucType, ptrInput));                             // SD-card disk group
#endif

    case DO_TELNET:
        fnDoTelnet(ucType, ptrInput);
        break;
#ifdef USE_PARAMETER_BLOCK                                               // {5}
    case DO_FLASH:
        fnDoFlash(ucType, ptrInput);
        break;
#endif

    case DO_IP:
        fnDoIP(ucType, ptrInput);
        break;

#if defined SERIAL_INTERFACE && defined DEMO_UART                        // {10}
    case DO_SERIAL:
        fnDoSerial(ucType, ptrInput);
        break;
#endif

    case DO_ADMIN:
        fnDoAdmin(ucType, ptrInput);
        break;

    case DO_SERVER:
        fnDoServer(ucType, ptrInput);;
        break;
    }
    return 0;
}

#ifdef SDCARD_SUPPORT                                                    // {17}
static void fnSendPrompt(void)
{
    if (ucMenu == MENU_HELP_DISK) {
        if ((ptr_utDirectory != 0) && (ptr_utDirectory->ptrDirectoryPath != 0)) {
            fnDebugMsg(ptr_utDirectory->ptrDirectoryPath);
        }
        fnDebugMsg(">");
    }
    else {
        fnDebugMsg("\r\n#");                                             // send prompt in command mode
    }
}
#endif


static int fnDoDebug(CHAR *ptr_input, unsigned short usInputLen, int iSource)
{
    const DEBUG_COMMAND *ptrCom = ucMenus[ucMenu].pMenu;
    int iEntries = ucMenus[ucMenu].ucMenuLength;
    CHAR *ptrNext;

    ptr_input[usInputLen-1] = 0;                                         // terminate input string

    if (iSource == SOURCE_SERIAL) {
        if ((usData_state == ES_NO_CONNECTION) || (usData_state & ES_SERIAL_SUSPENDED)) {
            return 0;                                                    // we can not treat commands
        }
    }
#ifdef USB_INTERFACE                                                     // {8}
    else if (iSource == SOURCE_USB) {
        if (usUSB_state == ES_NO_CONNECTION) {
            return 0;                                                    // we can not treat commands
        }
    }
#endif

    if (usInputLen) {
        while (iEntries--) {
            if ((ptrNext = fnCheckInput(ptr_input, ptrCom->strCommand)) != 0) {
                fnDebugMsg("\r\n");
                if (!fnDoCommand(ptrCom->ucDoFunction, ptrCom->ucType, ptrNext)) {
#ifdef SDCARD_SUPPORT                                                    // {17}
                    fnSendPrompt();
#else
                    fnDebugMsg("\r\n#");                                 // send prompt in command mode
#endif

                }
                if (iSource == SOURCE_SERIAL) {
                    fnGotoNextState(usData_state);                       // reinforce present state (to handle connection timers)
                }
                return 1;
            }
            ptrCom++;
        }
        iEntries = (sizeof(tSecretCommands) / sizeof(DEBUG_COMMAND));    // check whether we are to perform a secret command
        ptrCom = tSecretCommands;
        while (iEntries--) {
            if ((ptrNext = fnCheckInput(ptr_input, ptrCom->strCommand)) != 0) {
                if (!fnDoCommand(ptrCom->ucDoFunction, ptrCom->ucType, ptrNext)) {
                    fnDebugMsg("\r\n#");                                 // send prompt in command mode
                }
                if (iSource == SOURCE_SERIAL) {
                    fnGotoNextState(usData_state);                       // reinforce present state (to handle connection timers)
                }
                return 1;
            }
            ptrCom++;
        }
    }

    fnDebugMsg("\r\n??\r\n#");                                           // signal input non recognised and send prompt in command mode
    return 0;
}

static void fnLoginSuccess(void)
{
    if (ES_SERIAL_LOGIN == usData_state) {
        fnGotoNextState(ES_SERIAL_COMMAND_MODE);
    }
#ifdef USB_INTERFACE                                                     // {8}
    else if (ES_USB_LOGIN == usUSB_state) {
        usUSB_state = ES_USB_COMMAND_MODE;
        fnSetUSB_debug();                                                // enter USB channel as debug interface
    }
#endif
#ifdef USE_TELNET
    else {
        usTelnet_state = ES_NETWORK_COMMAND_MODE;
        fnTelnet(Telnet_socket, CLEAR_TEXT_ENTRY);                       // make sure telnet echos in clear text mode again
    }
#endif
    fnDoAdmin(DO_SHOW_CONFIG, 0);
    ucMenu = 0;                                                          // set main menu
    fnDisplayHelp(0);                                                    // display main menu
}


#ifdef USE_TELNET
// Telnet server
//
static int fnTELNETListener(USOCKET Socket, unsigned char ucEvent, unsigned char *ucIp_Data, unsigned short usPortLen)
{
    static QUEUE_HANDLE  BackupDebugPort;

    switch (ucEvent) {
    case TCP_EVENT_CONREQ:                                               // session request received on the TCP port
        if ((uMemcmp(temp_pars->temp_parameters.ucTrustedIP, cucNullMACIP, IPV4_LENGTH)) && (uMemcmp(temp_pars->temp_parameters.ucTrustedIP, ucIp_Data, IPV4_LENGTH))) {
            return APP_REJECT;                                           // we refuse the connection since we don't know the source
        }
        if (usData_state & (ES_CONNECTING_CLIENT | ES_SERIAL_COMMAND_MODE | ES_SERIAL_LOGIN)) { // if we have a serial command connection in progress, block a telnet connection
            return APP_REJECT;
        }
#ifdef USB_INTERFACE                                                     // {8}
        if (usUSB_state & (ES_USB_COMMAND_MODE | ES_USB_LOGIN)) {        // if we have a USB command connection in progress, block a telnet connection
            return APP_REJECT;
        }
#endif
        usTelnet_state = ES_CONNECTING_SERVER;
        break;

    case TCP_EVENT_CONNECTED:                                            // TCP connection has been established
        ucTelnetMode = 0;
                                                                         // we start negotiation of the telnet session and say that we will perform echo (only when not a binary TCP connection)
        usTelnet_state = ES_STARTING_COMMAND_MODE;                       // connection starts in command mode
        return fnTelnet(Telnet_socket, GOTO_ECHO_MODE);                  // try to set echo mode

    case TCP_EVENT_ABORT:
        break;

    case TCP_EVENT_CLOSE:
        usTelnet_state = ES_NO_CONNECTION;
    #ifdef SERIAL_INTERFACE
        ucPasswordState = PASSWORD_IDLE;
    #endif
        ucDebugCnt = 0;                                                  // invalidate any input in the debug input buffer
        break;

    case TCP_EVENT_CLOSED:
        if (DebugHandle == NETWORK_HANDLE) {
            DebugHandle = BackupDebugPort;                               // return original debug interface
        }
        usTelnet_state = ES_NO_CONNECTION;
    #ifdef SERIAL_INTERFACE
        ucPasswordState = PASSWORD_IDLE;
    #endif
        ucDebugCnt = 0;                                                  // invalidate any input in the debug input buffer
        break;

    case TCP_EVENT_ACK:
        if (usTelnet_state == ES_STARTING_COMMAND_MODE) {                // we have received an ack from our negotiation transmission(s) so we can welcome the user
            BackupDebugPort = DebugHandle;                               // save debug handle before telnet connection
            DebugHandle = NETWORK_HANDLE;                                // set the debug port to the network - we will send it over Telnet
            if (DO_PASSWORD_ENTRY == fnInitiateLogin(ES_NETWORK_LOGIN)) {// request password details before continuing
                fnTelnet(Telnet_socket, PASSWORD_ENTRY);                 // make sure telnet doesn't echo in clear text
            }
        }

        if (ES_TERMINATE_CONNECTION == usTelnet_state) {
            return APP_REQUEST_CLOSE;
        }
        break;

    case TCP_EVENT_DATA:
        if (fnCommandInput(ucIp_Data, usPortLen, SOURCE_NETWORK)) {      // collect data input and try to interpret if a LF is found
            return APP_SENT_DATA;                                        // we will always have sent an answer
        }
        break;
    }
    return APP_ACCEPT;                                                   // default is to accept a connection
}
#endif





extern CHAR *fnSkipWhiteSpace(CHAR *ptr_input)
{
    while ((*ptr_input == ' ') || (*ptr_input == TAB_KEY)) {
        ptr_input++;                                                     // remove white space
    }
    return (ptr_input);
}


extern int fnCommandInput(unsigned char *ptrData, unsigned short usLen, int iSource)
{
    #define MAX_DEBUG_IN 40
    int iReturn = 0;
    static CHAR cDebugIn[MAX_DEBUG_IN];

    while (usLen--) {
        if (*ptrData == DELETE_KEY) {
            if (ucDebugCnt) {
                ucDebugCnt--;                                            // we delete it from our local buffer but not from display
                fnDebugMsg((CHAR *)&cDeleteInput[1]);
            }
            continue;
        }
        if ((*ptrData == '\n') || (*ptrData == 0)) continue;             // some telnet clients send \r\n and some \r\0 - others only \r
        cDebugIn[ucDebugCnt] = *ptrData;
        if (*ptrData == '\r') {
            cDebugIn[ucDebugCnt] = '\n';
            fnSaveDebugHandle(iSource);                                  // save present debug handle and set appropriate
            fnDoDebug(cDebugIn, (unsigned char)(ucDebugCnt+1), iSource); // treat debug messages
            fnRestoreDebugHandle();                                      // return the present debug output
            iReturn = 1;                                                 // we will always send something so report that a transmission has been started
            ucDebugCnt = 0;
        }
        else {
            ucDebugCnt++;
        }
        ptrData++;
        if (ucDebugCnt >= MAX_DEBUG_IN) {
            ucDebugCnt = (MAX_DEBUG_IN - 1);
        }
    }
    return (iReturn);
}


extern int fnInitiateLogin(unsigned short usNextState)
{
    switch (usNextState) {                                               // {8}
#ifdef USB_INTERFACE
    case ES_USB_LOGIN:
        if ((usTelnet_state != ES_NO_CONNECTION) || (usData_state != ES_NO_CONNECTION)) { // can't login because already connected
            return TELNET_ON_LINE;
        }
        usUSB_state = usNextState;
        break;
#endif
#ifdef SERIAL_INTERFACE
    case ES_SERIAL_LOGIN:
        if ((usUSB_state != ES_NO_CONNECTION) || (usTelnet_state != ES_NO_CONNECTION)) { // can't login because already connected
            return TELNET_ON_LINE;
        }
        fnGotoNextState(usNextState);
        break;
#endif
    default:                                                             // network login
        usTelnet_state = ES_NETWORK_LOGIN;
        break;
    }
    fnLoginSuccess();                                                    // no login - go directly to connected state
    return DIRECT_LOGIN;
}

    #if defined SERIAL_INTERFACE
// Echo input or send start on password entry
//
extern void fnEchoInput(unsigned char *ucInputMessage, QUEUE_TRANSFER Length)
{
    if ((usTelnet_state != ES_NETWORK_LOGIN) && (ucPasswordState == TELNET_LOGIN)) {
        if (*ucInputMessage == DELETE_KEY) {
            fnWrite( SerialPortID, (unsigned char *)cDeleteInput, 3);
        }
        else if (*ucInputMessage == ':') {
            fnWrite( SerialPortID, ucInputMessage, 1);
        }
        else {                                                           // we want to hide password entry when logging in
            fnWrite( SerialPortID, (unsigned char *)"*", 1);             // send one star back (we assume human entry with one byte frames)
        }
    }
    else {
        fnWrite( SerialPortID, ucInputMessage, Length);                  // echo received character(s)
    }
}
    #endif
#else                                                                    // endif KEEP_DEBUG
void fnDebug(TTASKTABLE *ptrTaskTable)                                   // dummy task
{
}
#endif                                                                   // endif not KEEP_DEBUG



extern void fnConfigureTelnetServer(void)
{
#ifdef USE_TELNET
    if (temp_pars->temp_parameters.ucServers & ACTIVE_TELNET_SERVER) {
        Telnet_socket = fnStartTelnet(temp_pars->temp_parameters.usTelnetPort, (2*60), 0, 0, fnTELNETListener);   // 2 minute inactivity timeout
    }
    else {
        fnStopTelnet(Telnet_socket);
        usTelnet_state = ES_NO_CONNECTION;
    #ifdef SERIAL_INTERFACE
        ucPasswordState = PASSWORD_IDLE;
    #endif
    }
#endif
#ifdef TEST_CLIENT_SERVER
    fnTestTCPServer(TCP_TEST_SERVER_START);                              // start the test TCP server
#endif
}

extern void fnResetChanges(void)
{
    int iActions = 0;
    #define CHANGE_SERIAL_SETTINGS 0x01
    #define CHANGE_WEB_SERVER      0x02
    #define CHANGE_FTP_SERVER      0x04
    #define CHANGE_TELNET_SERVER   0x08
    if ((temp_pars->temp_parameters.usSerialMode != parameters->usSerialMode) ||  // check whether serial setting are being set back
        (temp_pars->temp_parameters.ucFlowHigh != parameters->ucFlowHigh) ||
        (temp_pars->temp_parameters.ucFlowLow != parameters->ucFlowLow) ||
        (temp_pars->temp_parameters.ucSerialSpeed != parameters->ucSerialSpeed))
    {
        iActions |= CHANGE_SERIAL_SETTINGS;
    }
    if ((temp_pars->temp_parameters.ucServers & (AUTHENTICATE_WEB_SERVER | ACTIVE_WEB_SERVER)) != (parameters->ucServers & (AUTHENTICATE_WEB_SERVER | ACTIVE_WEB_SERVER))) {
        iActions |= CHANGE_WEB_SERVER;
    }
    if ((temp_pars->temp_parameters.ucServers & (ACTIVE_FTP_SERVER)) != (parameters->ucServers & (ACTIVE_FTP_SERVER))) {
        iActions |= CHANGE_FTP_SERVER;
    }
    if ((temp_pars->temp_parameters.ucServers & (ACTIVE_TELNET_SERVER)) != (parameters->ucServers & (ACTIVE_TELNET_SERVER))) {
        iActions |= CHANGE_TELNET_SERVER;
    }
    fnGetOurParameters(1);                                               // get original parameters from FLASH, but preserve DHCP defined values

#if defined SERIAL_INTERFACE && defined DEMO_UART                        // {10}
    if (iActions & CHANGE_SERIAL_SETTINGS) {
        fnSetNewSerialMode(MODIFY_CONFIG);                               // return settings to interface
    }
#endif
    if (CHANGE_WEB_SERVER & iActions) {
#ifdef USE_HTTP
        fnConfigureAndStartWebServer();
#endif
    }
    if (CHANGE_FTP_SERVER & iActions) {
        fnConfigureFtpServer(FTP_TIMEOUT);                               // {3}
    }
    if (CHANGE_TELNET_SERVER & iActions) {
        fnConfigureTelnetServer();
    }
}


// Save port setup to flash
//
extern void fnSavePorts(void)
{
    unsigned char ucTempUserOutputs = temp_pars->temp_parameters.ucUserOutputs;
    unsigned char ucTempUserOutputValues = 0;
    unsigned short usTempUserDefinedOutputs = temp_pars->temp_parameters.usUserDefinedOutputs;
    CHAR cPort = '1';
    unsigned char ucBit = 1;

    while (cPort < '9') {
        if (fnPortState(cPort++)) {
            ucTempUserOutputValues |= ucBit;
        }
        ucBit <<= 1;
    }

    fnResetChanges();                                                    // ensure we don't save any unwanted changes

    temp_pars->temp_parameters.ucUserOutputs = ucTempUserOutputs;
    temp_pars->temp_parameters.ucUserOutputValues = ucTempUserOutputValues;
    temp_pars->temp_parameters.usUserDefinedOutputs = usTempUserDefinedOutputs;
    fnSaveNewPars(SAVE_NEW_PARAMETERS);                                  // save these settings as default
}


// Check the present state of a particular port
//
extern int fnPortState(CHAR cPortBit)
{
#ifdef DEMO9S12NE64
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;
    if (PTG & ucBit) {
        return 1;
    }
    return 0;
#endif
#if defined _HW_SAM7X
    #ifdef _HW_SAM7S
    return ((PIO_PDSR_A & (FIRST_USER_PORT << (cPortBit - '1'))) != 0);  // return non zero if the port is reading a '1'
    #else
    return ((PIO_PDSR_B & (FIRST_USER_PORT << (cPortBit - '1'))) != 0);  // return non zero if the port is reading a '1'
    #endif
#endif
#if defined AVR32_EVK1105 || defined AVR32_EVK1101 || defined AVR32_EVK1100  // {16}
    return ((PVR_1 & (FIRST_USER_PORT << (cPortBit - '1'))) != 0);       // return non zero if the port is reading a '1'
#endif
#if (defined _M5223X || defined _FLEXIS32) && !defined _M520X            // {1}{23}
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;

    if (ucBit & DDRTC) {
        if (PORTTC & ucBit) {
            return 1;                                                    // display as selected if output driven to '1'
        }
    }
    else {
        if (PORTIN_SETTC & ucBit) {
            return 1;                                                    // display as selected if input pulled to '1'
        }
    }
    return 0;
#endif
#if defined _STR91XF || (defined _LPC23XX && defined KEIL_MCB2300) || defined _LM3SXXXX
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;
    if (DEMO_LED_PORT & ucBit) {
        return 1;
    }
    return 0;
#elif defined _STM32
    switch (cPortBit) {
    case '1':
        if (DEMO_INPUT_PORT & DEMO_LED_1) {
            return 1;
        }
        break;
    case '2':
        if (DEMO_INPUT_PORT & DEMO_LED_2) {
            return 1;
        }
        break;
    case '3':
        if (DEMO_INPUT_PORT & DEMO_LED_3) {
            return 1;
        }
        break;
    default:
        break;
    }
    return 0;
#elif defined _LPC23XX || defined _LPC17XX
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;
    if ((DEMO_LED_PORT >> FIRST_USER_PORT) & ucBit) {
        return 1;
    }
    return 0;
#endif
}


// Check whether the port is configured as an input
//
extern int fnPortInputConfig(CHAR cPortBit)
{
#ifdef DEMO9S12NE64
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;
    if (DDRG & ucBit) {
        return 0;
    }
    return 1;
#endif
#if defined _HW_SAM7X
    #ifdef _HW_SAM7S
    return ((PIO_OSR_A & (FIRST_USER_PORT << (cPortBit - '1'))) == 0);   // return zero if the port is presently programmed as output
    #else
    return ((PIO_OSR_B & (FIRST_USER_PORT << (cPortBit - '1'))) == 0);   // return zero if the port is presently programmed as output
    #endif
#endif
#if defined AVR32_EVK1105 || defined AVR32_EVK1101 || defined AVR32_EVK1100  // {16}
    return ((ODER_1 & (FIRST_USER_PORT << (cPortBit - '1'))) == 0);      // return zero if the port is presently programmed as output
#endif
#if (defined _M5223X || defined _FLEXIS32) && !defined _M520X            // {23}
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;
    if (DDRTC & ucBit) {
        return 0;
    }
    return 1;
#endif
#if defined _STR91XF || (defined _LPC23XX && defined KEIL_MCB2300) || defined _LM3SXXXX
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;
    if (DEMO_LED_DDR & ucBit) {
        return 0;
    }
    return 1;
#elif defined _STM32                                                     // this device doesn't have a conventional DDR register
    switch (cPortBit) {
    case '1':
        if (GPIOD_CRL & 0x30000) {
            return 1;                                                    // configured as output
        }
        break;
    case '2':
        if (GPIOD_CRL & 0x3000) {
            return 1;
        }
        break;
    case '3':
        if (GPIOD_CRL & 0x30000000) {
            return 1;
        }
        break;
    default:
        break;
    }
    return 0;
#elif defined _LPC23XX || defined _LPC17XX
    unsigned char ucBit = 0x01;
    cPortBit -= '1';
    ucBit <<= cPortBit;
    if ((DEMO_LED_DDR>>FIRST_USER_PORT) & ucBit) {
        return 0;
    }
    return 1;
#endif
}

extern int fnConfigPort(CHAR cPortBit, CHAR cType)
{
    cPortBit -= '0';                                                     // select bit offset

    switch (cType) {
    case 'd':                                                            // default use for 0..4
    case 'i':                                                            // port to be input
#ifdef DEMO9S12NE64
        if (cPortBit > 3) break;
        DDRG &= ~(1<<cPortBit);                                          // set port bit to input
        temp_pars->temp_parameters.ucUserOutputs = DDRG;
#endif
#if defined _HW_SAM7S
        if (cPortBit < 4) {
            PIO_ODR_A = (FIRST_USER_PORT << cPortBit);       _SIM_PORTS; // port bits are assumed to be next to another
            temp_pars->temp_parameters.ucUserOutputs = (unsigned char)((PIO_OSR_A & DEMO_USER_PORTS) >> SHIFT_PORT_TO_BYTE);
        }
#elif defined _HW_SAM7X
        if (cPortBit < 4) {
            PIO_ODR_B = (FIRST_USER_PORT << cPortBit);       _SIM_PORTS; // port bits are assumed to be next to another
            temp_pars->temp_parameters.ucUserOutputs = (unsigned char)((PIO_OSR_B & DEMO_USER_PORTS) >> SHIFT_PORT_TO_BYTE);
        }
#endif
#ifdef AVR32_EVK1105
        if (cPortBit < 4) {
            ODERC_1 = (FIRST_USER_PORT << cPortBit);                     // port bits are assumed to be next to another on port B
    #ifdef _WINDOWS
            fnSimPorts();                                                // ensure the simulator updates its port states
    #endif
            temp_pars->temp_parameters.ucUserOutputs = (unsigned char)((ODER_1 & DEMO_USER_PORTS) >> SHIFT_PORT_TO_BYTE);
        }
#endif
#if (defined _M5223X || defined _FLEXIS32) && !defined _M520X            // {1}{23}
        if (cPortBit > 3) {
            break;
        }
        DDRTC &= ~(1<<cPortBit);                                         // set port bit to input
        temp_pars->temp_parameters.ucUserOutputs = DDRTC;
#endif
#if defined _LPC23XX || defined _LPC17XX
        if (cPortBit > 3) {
            break;
        }
        DEMO_LED_DDR &= ~((1 << cPortBit) << FIRST_USER_PORT);           // set port bit to input
        temp_pars->temp_parameters.ucUserOutputs = (unsigned char)(DEMO_LED_DDR>>FIRST_USER_PORT);
#endif
#ifdef _STR91XF
        if (cPortBit > 3) {
            break;
        }
        DEMO_LED_CONFIG &= ~(MASK_FUNCTION << (2*cPortBit));             // mask function
        DEMO_LED_DDR &= ~(1 << cPortBit);                                // set port bit to input
        temp_pars->temp_parameters.ucUserOutputs = (unsigned char)DEMO_LED_DDR;
#endif
#ifdef _STM32
        switch (cPortBit) {                                              // 3 user ports available
        case 7:
            _CONFIG_PORT_INPUT(D, (unsigned char)LED1, (LED1 << 16));    // configure as input with pull-up
            temp_pars->temp_parameters.ucUserOutputs &= ~LED1;           // set present bit as input
            break;
        case 4:
            _CONFIG_PORT_INPUT(D, (unsigned char)LED4, (LED4 << 16));    // configure as input with pull-up
            temp_pars->temp_parameters.ucUserOutputs &= ~LED4;           // set present bit as input
            break;
        case 3:
            _CONFIG_PORT_INPUT(D, (unsigned char)LED3, (LED3 << 16));    // configure as input with pull-up
            temp_pars->temp_parameters.ucUserOutputs &= ~LED3;           // set present bit as input
            break;
        default:
            break;
        }
#endif
#ifdef _LM3SXXXX
    #if defined EK_LM3S3748 && (defined GLCD_COLOR || defined CGLCD_GLCD_MODE) // {15}
        if (cPortBit == 1) {                                             // don't disturn CGLCD backlight
            break;
        }
    #endif
        if (cPortBit > 3) {                                              // 4 user ports available
            break;
        }
        DEMO_LED_DDR &= ~(1 << cPortBit);                                // set port bit to input
        DEMO_DIGITAL_ENABLE |= (1 << cPortBit);                          // connect for digital use
        temp_pars->temp_parameters.ucUserOutputs = (unsigned char)DEMO_LED_DDR;
#endif
        break;

    case 'o':                                                            // port to be output
#ifdef DEMO9S12NE64
        if (cPortBit > 3) {
            break;
        }
        DDRG |= (1 << cPortBit);                                         // set port bit to output
        temp_pars->temp_parameters.ucUserOutputs = DDRG;
#endif
#if defined _HW_SAM7S
        if (cPortBit < 4) {
            PIO_OER_A = (FIRST_USER_PORT << cPortBit);       _SIM_PORTS; // port bits are assumed to be next to another
            temp_pars->temp_parameters.ucUserOutputs = (unsigned char)((PIO_OSR_A & DEMO_USER_PORTS) >> SHIFT_PORT_TO_BYTE);
        }
#elif defined _HW_SAM7X
        if (cPortBit < 4) {
            PIO_OER_B = (FIRST_USER_PORT << cPortBit);       _SIM_PORTS; // port bits are assumed to be next to another
            temp_pars->temp_parameters.ucUserOutputs = (unsigned char)((PIO_OSR_B & DEMO_USER_PORTS) >> SHIFT_PORT_TO_BYTE);
        }
#endif
#ifdef AVR32_EVK1105
        if (cPortBit < 4) {
            ODERS_1 = (FIRST_USER_PORT << cPortBit);                     // port bits are assumed to be next to another on port B
    #ifdef _WINDOWS
            fnSimPorts();                                                // ensure the simulator updates its port states
    #endif
            temp_pars->temp_parameters.ucUserOutputs = (unsigned char)((ODER_1 & DEMO_USER_PORTS) >> SHIFT_PORT_TO_BYTE);
        }
#endif
#if (defined _M5223X || defined _FLEXIS32) && !defined _M520X            // {1}{23}
        if (cPortBit > 3) {
            break;
        }
        DDRTC |= (1 << cPortBit);                                        // set port bit to output
        temp_pars->temp_parameters.ucUserOutputs = DDRTC;
#endif
#if defined _LPC23XX || defined _LPC17XX
        if (cPortBit > 3) {
            break;
        }
        DEMO_LED_DDR |= ((1 << cPortBit) << FIRST_USER_PORT);            // set port bit to output
        temp_pars->temp_parameters.ucUserOutputs = (unsigned char)(DEMO_LED_DDR >> FIRST_USER_PORT);
#endif
#ifdef _STR91XF
        if (cPortBit > 3) {
            break;
        }
        DEMO_LED_CONFIG &= ~(MASK_FUNCTION << (2*cPortBit));             // mask function
        DEMO_LED_CONFIG |= (PORT_OUT_FUNCTION << (2*cPortBit));          // set to output function
        DEMO_LED_DDR |= (1 << cPortBit);                                 // set port bit to output
        temp_pars->temp_parameters.ucUserOutputs = (unsigned char)DEMO_LED_DDR;
#endif
#ifdef _STM32
        switch (cPortBit) {                                              // 3 user ports available
        case 7:
            _CONFIG_PORT_OUTPUT(D, (unsigned char)LED1, (OUTPUT_PUSH_PULL | OUTPUT_MEDIUM)) // configure as medium speed output with push-pull output
            temp_pars->temp_parameters.ucUserOutputs |= LED1;            // set present bit as output
            break;
        case 4:
            _CONFIG_PORT_OUTPUT(D, (unsigned char)LED4, (OUTPUT_PUSH_PULL | OUTPUT_MEDIUM)) // configure as medium speed output with push-pull output
            temp_pars->temp_parameters.ucUserOutputs |= LED4;            // set present bit as output
            break;
        case 3:
            _CONFIG_PORT_OUTPUT(D, (unsigned char)LED3, (OUTPUT_PUSH_PULL | OUTPUT_MEDIUM)) // configure as medium speed output with push-pull output
            temp_pars->temp_parameters.ucUserOutputs |= LED3;            // set present bit as output
            break;
        default:
            break;
        }
#endif
#ifdef _LM3SXXXX
        if (cPortBit > 3) {                                              // 4 user ports available
            break;
        }
        DEMO_LED_DDR |= (1 << cPortBit);                                 // set port bit to output
        DEMO_DIGITAL_ENABLE |= (1 << cPortBit);                          // connect for digital use
        temp_pars->temp_parameters.ucUserOutputs = (unsigned char)DEMO_LED_DDR;
#endif
        break;

    case 'a':                                                            // Port to be analog
#ifdef DEMO9S12NE64
        ATDCTL4 = 0xf3;                                                  // 8 bit mode suitable for our bus speed
        ATDCTL2 = ADPU;                                                  // power up the ADC
        ATDCTL5 = SCAN;                                                  // start scanning
#endif
        break;

    default:
        return -1;
    }
    return 0;
}




#if !defined _M520X                                                      // {23}
#if ((defined (_M5223X) || defined _FLEXIS32) && !(defined (SUPPORT_KEY_SCAN))) || defined (_HW_NE64) || defined (_STR91XF) || defined (_LM3SXXXX) || defined _STM32
// Set all port bits to new state
// These processor examples uses a register which holds the present value. It is modifed and written back again.
static void fnSetUserPortOut(unsigned short usPortOutputs)
{
    unsigned short usBit = 0x0001;

    while (usBit) {
        if (usPortOutputs & usBit) {
            switch (usBit) {
            case 0x0001:
                USER_PORT_1 |= USER_PORT_1_BIT;
                break;
            case 0x0002:
                USER_PORT_2 |= USER_PORT_2_BIT;
                break;
            case 0x0004:
                USER_PORT_3 |= USER_PORT_3_BIT;
                break;
            case 0x0008:
                USER_PORT_4 |= USER_PORT_4_BIT;
                break;
    #ifndef _HW_NE64
            case 0x0010:
                USER_PORT_5 |= USER_PORT_5_BIT;
                break;
            case 0x0020:
                USER_PORT_6 |= USER_PORT_6_BIT;
                break;
            case 0x0040:
                USER_PORT_7 |= USER_PORT_7_BIT;
                break;
            case 0x0080:
                USER_PORT_8 |= USER_PORT_8_BIT;
                break;
            case 0x0100:
                USER_PORT_9 |= USER_PORT_9_BIT;
                break;
            case 0x0200:
                USER_PORT_10 |= USER_PORT_10_BIT;
                break;
            case 0x0400:
                USER_PORT_11 |= USER_PORT_11_BIT;
                break;
            case 0x0800:
                USER_PORT_12 |= USER_PORT_12_BIT;
                break;
            case 0x1000:
                USER_PORT_13 |= USER_PORT_13_BIT;
                break;
            case 0x2000:
                USER_PORT_14 |= USER_PORT_14_BIT;
                break;
            case 0x4000:
                USER_PORT_15 |= USER_PORT_15_BIT;
                break;
            case 0x8000:
                USER_PORT_16 |= USER_PORT_16_BIT;
                break;
    #endif
            }
        }
        else {
            switch (usBit) {
            case 0x0001:
                USER_PORT_1 &= ~USER_PORT_1_BIT;
                break;
            case 0x0002:
                USER_PORT_2 &= ~USER_PORT_2_BIT;
                break;
            case 0x0004:
                USER_PORT_3 &= ~USER_PORT_3_BIT;
                break;
            case 0x0008:
                USER_PORT_4 &= ~USER_PORT_4_BIT;
                break;
    #ifndef _HW_NE64
            case 0x0010:
                USER_PORT_5 &= ~USER_PORT_5_BIT;
                break;
            case 0x0020:
                USER_PORT_6 &= ~USER_PORT_6_BIT;
                break;
            case 0x0040:
                USER_PORT_7 &= ~USER_PORT_7_BIT;
                break;
            case 0x0080:
                USER_PORT_8 &= ~USER_PORT_8_BIT;
                break;
            case 0x0100:
                USER_PORT_9 &= ~USER_PORT_9_BIT;
                break;
            case 0x0200:
                USER_PORT_10 &= ~USER_PORT_10_BIT;
                break;
            case 0x0400:
                USER_PORT_11 &= ~USER_PORT_11_BIT;
                break;
            case 0x0800:
                USER_PORT_12 &= ~USER_PORT_12_BIT;
                break;
            case 0x1000:
                USER_PORT_13 &= ~USER_PORT_13_BIT;
                break;
            case 0x2000:
                USER_PORT_14 &= ~USER_PORT_14_BIT;
                break;
            case 0x4000:
                USER_PORT_15 &= ~USER_PORT_15_BIT;
                break;
            case 0x8000:
                USER_PORT_16 &= ~USER_PORT_16_BIT;
                break;
    #endif
            }
        }
        usBit <<= 1;
    }
    temp_pars->temp_parameters.usUserDefinedOutputs = usPortOutputs;     // backup the new setting
    #ifdef _WINDOWS
    fnSimPorts();                                                        // ensure the simulator updates its port states
    #endif
}
#endif
#endif

#if defined (_HW_SAM7X) || (defined (_LPC23XX) && !defined (_LPC21XX)) || defined (_HW_AVR32) || defined (_LPC17XX) // {11}
static void fnSetPortBit(unsigned short usBit, int iSetClr)
{
    if (iSetClr) {
        switch (usBit) {
        case 0x0001:
            USER_SET_PORT_1 = USER_PORT_1_BIT;
            break;
        case 0x0002:
            USER_SET_PORT_2 = USER_PORT_2_BIT;
            break;
        case 0x0004:
            USER_SET_PORT_3 = USER_PORT_3_BIT;
            break;
        case 0x0008:
            USER_SET_PORT_4 = USER_PORT_4_BIT;
            break;
        case 0x0010:
            USER_SET_PORT_5 = USER_PORT_5_BIT;
            break;
        case 0x0020:
            USER_SET_PORT_6 = USER_PORT_6_BIT;
            break;
        case 0x0040:
            USER_SET_PORT_7 = USER_PORT_7_BIT;
            break;
        case 0x0080:
            USER_SET_PORT_8 = USER_PORT_8_BIT;
            break;
        case 0x0100:
            USER_SET_PORT_9 = USER_PORT_9_BIT;
            break;
        case 0x0200:
            USER_SET_PORT_10 = USER_PORT_10_BIT;
            break;
        case 0x0400:
            USER_SET_PORT_11 = USER_PORT_11_BIT;
            break;
        case 0x0800:
            USER_SET_PORT_12 = USER_PORT_12_BIT;
            break;
        case 0x1000:
            USER_SET_PORT_13 = USER_PORT_13_BIT;
            break;
        case 0x2000:
            USER_SET_PORT_14 = USER_PORT_14_BIT;
            break;
        case 0x4000:
            USER_SET_PORT_15 = USER_PORT_15_BIT;
            break;
        case 0x8000:
            USER_SET_PORT_16 = USER_PORT_16_BIT;
            break;
        }
    }
    else {
        switch (usBit) {
        case 0x0001:
            USER_CLEAR_PORT_1 = USER_PORT_1_BIT;
            break;
        case 0x0002:
            USER_CLEAR_PORT_2 = USER_PORT_2_BIT;
            break;
        case 0x0004:
            USER_CLEAR_PORT_3 = USER_PORT_3_BIT;
            break;
        case 0x0008:
            USER_CLEAR_PORT_4 = USER_PORT_4_BIT;
            break;
        case 0x0010:
            USER_CLEAR_PORT_5 = USER_PORT_5_BIT;
            break;
        case 0x0020:
            USER_CLEAR_PORT_6 = USER_PORT_6_BIT;
            break;
        case 0x0040:
            USER_CLEAR_PORT_7 = USER_PORT_7_BIT;
            break;
        case 0x0080:
            USER_CLEAR_PORT_8 = USER_PORT_8_BIT;
            break;
        case 0x0100:
            USER_CLEAR_PORT_9 = USER_PORT_9_BIT;
            break;
        case 0x0200:
            USER_CLEAR_PORT_10 = USER_PORT_10_BIT;
            break;
        case 0x0400:
            USER_CLEAR_PORT_11 = USER_PORT_11_BIT;
            break;
        case 0x0800:
            USER_CLEAR_PORT_12 = USER_PORT_12_BIT;
            break;
        case 0x1000:
            USER_CLEAR_PORT_13 = USER_PORT_13_BIT;
            break;
        case 0x2000:
            USER_CLEAR_PORT_14 = USER_PORT_14_BIT;
            break;
        case 0x4000:
            USER_CLEAR_PORT_15 = USER_PORT_15_BIT;
            break;
        case 0x8000:
            USER_CLEAR_PORT_16 = USER_PORT_16_BIT;
            break;
        }
    }
    #ifdef _WINDOWS
    fnSimPorts();                                                        // ensure the simulator updates its port states
    #endif
}

// Set all port bits to new state
// The ARM processors tend to use set and clear registers to perform the job
//
static void fnSetUserPortOut(unsigned short usPortOutputs)
{
    unsigned short usBit = 0x0001;
    #if defined _HW_SAM7X
    POWER_UP(PIOA | PIOB);                                               // {9} enable peripheral clock to PIO channel A and B
    #endif

    while (usBit) {
        fnSetPortBit(usBit, usPortOutputs & usBit);
        usBit <<= 1;
    #ifdef _WINDOWS
        fnSimPorts();                                                    // ensure the simulator updates its port states
    #endif
    }
    temp_pars->temp_parameters.usUserDefinedOutputs = usPortOutputs;     // backup the new setting
}

#endif
// Request present level of an output
//
extern int fnUserPortState(CHAR cPortBit)
{
#if !defined _M520X                                                      // {23}
    switch (cPortBit) {
    case 'a':
        return ((USER_PORT_1 & USER_PORT_1_BIT) != 0);
    case 'b':
        return ((USER_PORT_2 & USER_PORT_2_BIT) != 0);
    case 'c':
        return ((USER_PORT_3 & USER_PORT_3_BIT) != 0);
    case 'd':
        return ((USER_PORT_4 & USER_PORT_4_BIT) != 0);
#ifndef _HW_NE64
    case 'e':
        return ((USER_PORT_5 & USER_PORT_5_BIT) != 0);
    case 'f':
        return ((USER_PORT_6 & USER_PORT_6_BIT) != 0);
    case 'g':
        return ((USER_PORT_7 & USER_PORT_7_BIT) != 0);
    case 'h':
        return ((USER_PORT_8 & USER_PORT_8_BIT) != 0);
    case 'i':
        return ((USER_PORT_9 & USER_PORT_9_BIT) != 0);
    case 'j':
        return ((USER_PORT_10 & USER_PORT_10_BIT) != 0);
    case 'k':
        return ((USER_PORT_11 & USER_PORT_11_BIT) != 0);
    case 'l':
        return ((USER_PORT_12 & USER_PORT_12_BIT) != 0);
    case 'm':
        return ((USER_PORT_13 & USER_PORT_13_BIT) != 0);
    case 'n':
        return ((USER_PORT_14 & USER_PORT_14_BIT) != 0);
    case 'o':
        return ((USER_PORT_15 & USER_PORT_15_BIT) != 0);
    case 'p':
        return ((USER_PORT_16 & USER_PORT_16_BIT) != 0);
#endif
    default:
        break;
    }
#endif
    return 0;
}
#if !defined _M520X
#if ((defined (_M5223X) || defined _FLEXIS32) && !(defined (SUPPORT_KEY_SCAN))) || defined (_HW_NE64)|| defined (_STR91XF) || defined (_LM3SXXXX) || (defined (_LPC23XX) && !defined (_LPC21XX)) || defined (_LPC17XX) || defined (_STM32)
// Configure port bit as output (a..p are port bit references)
//
static int fnConfigOutputPort(CHAR cPortBit)
{
    switch (cPortBit) {
    case 'a':
    #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_1();
        CONFIG_USER_PORT_1();
    #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_1();
    #endif
    #ifndef _STM32
        USER_DDR_1 |= USER_PORT_1_BIT;
    #endif
        break;
    case 'b':
    #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_2();
        CONFIG_USER_PORT_2();
    #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_2();
    #endif
    #ifndef _STM32
        USER_DDR_2 |= USER_PORT_2_BIT;
        break;
    #endif
    case 'c':
    #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_3();
        CONFIG_USER_PORT_3();
    #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_3();
    #endif
    #ifndef _STM32
        USER_DDR_3 |= USER_PORT_3_BIT;
    #endif
        break;
    case 'd':
    #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_4();
        CONFIG_USER_PORT_4();
    #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_4();
    #endif
    #ifndef _STM32
        USER_DDR_4 |= USER_PORT_4_BIT;
        break;
    #endif
    #ifndef _HW_NE64
    case 'e':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_5();
        CONFIG_USER_PORT_5();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_5();
        #endif
        #ifndef _STM32
        USER_DDR_5 |= USER_PORT_5_BIT;
        #endif
        break;
    case 'f':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_6();
        CONFIG_USER_PORT_6();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_6();
        #endif
        #ifndef _STM32
        USER_DDR_6 |= USER_PORT_6_BIT;
        #endif
        break;
    case 'g':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_7();
        CONFIG_USER_PORT_7();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_7();
        #endif
        #ifndef _STM32
        USER_DDR_7 |= USER_PORT_7_BIT;
        #endif
        break;
    case 'h':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_8();
        CONFIG_USER_PORT_8();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_8();
        #endif
        #ifndef _STM32
        USER_DDR_8 |= USER_PORT_8_BIT;
        #endif
        break;
    case 'i':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_9();
        CONFIG_USER_PORT_9();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_9();
        #endif
        #ifndef _STM32
        USER_DDR_9 |= USER_PORT_9_BIT;
        #endif
        break;
    case 'j':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_10();
        CONFIG_USER_PORT_10();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_10();
        #endif
        #ifndef _STM32
        USER_DDR_10 |= USER_PORT_10_BIT;
        #endif
        break;
    case 'k':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_11();
        CONFIG_USER_PORT_11();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_11();
        #endif
        #ifndef _STM32
        USER_DDR_11 |= USER_PORT_11_BIT;
        #endif
        break;
    case 'l':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_12();
        CONFIG_USER_PORT_12();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_12();
        #endif
        #ifndef _STM32
        USER_DDR_12 |= USER_PORT_12_BIT;
        #endif
        break;
    case 'm':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_13();
        CONFIG_USER_PORT_13();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_13();
        #endif
        #ifndef _STM32
        USER_DDR_13 |= USER_PORT_13_BIT;
        #endif
        break;
    case 'n':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_14();
        CONFIG_USER_PORT_14();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_14();
        #endif
        #ifndef _STM32
        USER_DDR_14 |= USER_PORT_14_BIT;
        #endif
        break;
    case 'o':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_15();
        CONFIG_USER_PORT_15();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_15();
        #endif
        #ifndef _STM32
        USER_DDR_15 |= USER_PORT_15_BIT;
        #endif
        break;
    case 'p':
        #if defined (_STR91XF) || defined (_STM32)
        ENABLE_USER_PORT_16();
        CONFIG_USER_PORT_16();
        #elif defined (_LM3SXXXX)
        ENABLE_USER_PORT_16();
        #endif
        #ifndef _STM32
        USER_DDR_16 |= USER_PORT_16_BIT;
        #endif
        break;
    #endif
    default:
        return 1;
    }
    #ifdef _WINDOWS
    fnSimPorts();                                                        // ensure the simulator updates its port states
    #endif
    return 0;
}
#endif
#endif

#if defined _HW_SAM7X || defined (_HW_AVR32)                             // {11}
// Configure port bit as output (a..p are port bit references)
//
static int fnConfigOutputPort(CHAR cPortBit)
{
    switch (cPortBit) {
    case 'a':
        USER_DDR_1 = USER_PORT_1_BIT;
        break;
    case 'b':
        USER_DDR_2 = USER_PORT_2_BIT;
        break;
    case 'c':
        USER_DDR_3 = USER_PORT_3_BIT;
        break;
    case 'd':
        USER_DDR_4 = USER_PORT_4_BIT;
        break;
    case 'e':
        USER_DDR_5 = USER_PORT_5_BIT;
        break;
    case 'f':
        USER_DDR_6 = USER_PORT_6_BIT;
        break;
    case 'g':
        USER_DDR_7 = USER_PORT_7_BIT;
        break;
    case 'h':
        USER_DDR_8 = USER_PORT_8_BIT;
        break;
    case 'i':
        USER_DDR_9 = USER_PORT_9_BIT;
        break;
    case 'j':
        USER_DDR_10 = USER_PORT_10_BIT;
        break;
    case 'k':
        USER_DDR_11 = USER_PORT_11_BIT;
        break;
    case 'l':
        USER_DDR_12 = USER_PORT_12_BIT;
        break;
    case 'm':
        USER_DDR_13 = USER_PORT_13_BIT;
        break;
    case 'n':
        USER_DDR_14 = USER_PORT_14_BIT;
        break;
    case 'o':
        USER_DDR_15 = USER_PORT_15_BIT;
        break;
    case 'p':
        USER_DDR_16 = USER_PORT_16_BIT;
        break;
    default:
        return 1;
    }
                                                             _SIM_PORTS; // ensure the simulator updates its port states
    return 0;
}
#endif

#if defined (_M5223X) || defined _FLEXIS32 || defined (_HW_NE64) || defined (_STR91XF) || defined (_LM3SXXXX) || (defined _STM32)
// Toggle the state of an output port
//
extern int fnTogglePortOut(CHAR cPortBit)
{
    #if !defined _M520X                                                  // {23}
    switch (cPortBit) {
    case 'a':
        USER_PORT_1 ^= USER_PORT_1_BIT;
        break;
    case 'b':
        USER_PORT_2 ^= USER_PORT_2_BIT;
        break;
    case 'c':
        USER_PORT_3 ^= USER_PORT_3_BIT;
        break;
    case 'd':
        USER_PORT_4 ^= USER_PORT_4_BIT;
        break;
    #ifndef _HW_NE64
    case 'e':
        USER_PORT_5 ^= USER_PORT_5_BIT;
        break;
    case 'f':
        USER_PORT_6 ^= USER_PORT_6_BIT;
        break;
    case 'g':
        USER_PORT_7 ^= USER_PORT_7_BIT;
        break;
    case 'h':
        USER_PORT_8 ^= USER_PORT_8_BIT;
        break;
    case 'i':
        USER_PORT_9 ^= USER_PORT_9_BIT;
        break;
    case 'j':
        USER_PORT_10 ^= USER_PORT_10_BIT;
        break;
    case 'k':
        USER_PORT_11 ^= USER_PORT_11_BIT;
        break;
    case 'l':
        USER_PORT_12 ^= USER_PORT_12_BIT;
        break;
    case 'm':
        USER_PORT_13 ^= USER_PORT_13_BIT;
        break;
    case 'n':
        USER_PORT_14 ^= USER_PORT_14_BIT;
        break;
    case 'o':
        USER_PORT_15 ^= USER_PORT_15_BIT;
        break;
    case 'p':
        USER_PORT_16 ^= USER_PORT_16_BIT;
        break;
    #endif
    default:
        break;
    }
    temp_pars->temp_parameters.usUserDefinedOutputs ^= (0x0001 << (cPortBit - 'a'));
    #endif
    return 0;
}
#endif

#if defined (_HW_SAM7X) || (defined (_LPC23XX) && !defined (_LPC21XX)) || defined (_HW_AVR32) || defined _LPC17XX // {11}
// Toggle the state of an output port
//
extern int fnTogglePortOut(CHAR cPortBit)
{
    unsigned short usNewBit = (0x0001 << (cPortBit - 'a'));

    if (fnUserPortState(cPortBit)) {
        fnSetPortBit(usNewBit, 0);
    }
    else {
        fnSetPortBit(usNewBit, 1);
    }

    temp_pars->temp_parameters.usUserDefinedOutputs ^= usNewBit;
    return 0;
}
#endif

// Initialise ports to input/output
//
extern void fnInitialisePorts(void)
{
    CHAR sPort = '0';
    CHAR cType;
    unsigned char ucBit = 1;
    unsigned char ucUserOutputs = temp_pars->temp_parameters.ucUserOutputs;

    temp_pars->temp_parameters.ucUserOutputValues &= ucUserOutputs;      // clear inputs
    fnSetPortOut(temp_pars->temp_parameters.ucUserOutputValues, 1);      // configure all user outputs to default states

    while (sPort < '8') {                                                // up to 8 demo port bits
        cType = 'o';
        if (!(ucUserOutputs & ucBit)) {
            cType = 'i';
        }
        fnConfigPort(sPort++, cType);                                    // configure port as input or output
        ucBit <<= 1;
    }
#if !defined _M520X                                                      // {23}
                                                                         // user defined outputs used by keypad so they may not be set here
    #if ((defined (_M5223X) || defined _FLEXIS32) && !(defined (SUPPORT_KEY_SCAN))) || defined (_HW_SAM7X) || defined (_HW_NE64) || defined (_STR91XF) || (defined (_LM3SXXXX) && !defined _LM3S10X) || (defined (_LPC23XX) && !defined (_LPC21XX)) || defined (_HW_AVR32) || defined (_LPC17XX) || defined _STM32 // {11}
        #ifdef _LM3SXXXX
    POWER_UP_USER_PORTS();
        #endif
                                                                         // initialise user ports as outputs with default values
    sPort = 'a';                                                         // first user output port
    fnSetUserPortOut(temp_pars->temp_parameters.usUserDefinedOutputs);   // configure all user outputs to default states
    while (sPort < 'q') {                                                // until last port
        fnConfigOutputPort(sPort++);                                     // configure port bits as outputs
    }
    #endif
#endif
}

extern void fnSetPortOut(unsigned char ucPortOutputs, int iInitialisation)
{
#ifdef DEMO9S12NE64
    unsigned char ucPortValue = (PTG & ~0x0f);                           // {4}
    ucPortValue |= ucPortOutputs;
    PTG = (ucPortValue);
    temp_pars->temp_parameters.ucUserOutputValues = ucPortOutputs;
#endif
#if defined _HW_SAM7S
    unsigned long ulChange = ((ucPortOutputs << SHIFT_PORT_TO_BYTE) & DEMO_USER_PORTS);
    POWER_UP(PIOA);                                                      // ensure port has clocks enabled
    PIO_SODR_A = (ulChange & DEMO_USER_PORTS);
    PIO_CODR_A = (~ulChange & DEMO_USER_PORTS);                          _SIM_PORTS;
#elif defined _HW_SAM7X
    unsigned long ulChange = ((ucPortOutputs << SHIFT_PORT_TO_BYTE) & DEMO_USER_PORTS);
    POWER_UP(PIOB);                                                      // {9} ensure port has clocks enabled
    PIO_SODR_B = (ulChange & DEMO_USER_PORTS);
    PIO_CODR_B = (~ulChange & DEMO_USER_PORTS);                          _SIM_PORTS;
#endif
#ifdef AVR32_EVK1105
    unsigned long ulChange = ((ucPortOutputs << SHIFT_PORT_TO_BYTE) & DEMO_USER_PORTS);
    OVRS_1 = (ulChange & DEMO_USER_PORTS);
    OVRC_1 = (~ulChange & DEMO_USER_PORTS);
    #ifdef _WINDOWS
    fnSimPorts();                                                        // ensure the simulator updates its port states
    #endif
#endif
#if (defined _M5223X || defined _FLEXIS32) && !defined _M520X            // {1}{23}
    if (!iInitialisation) {
        ucPortOutputs &= DDRTC;                                          // remove non-output bits
    }
    PORTTC = ucPortOutputs;                                              // set port values
    temp_pars->temp_parameters.ucUserOutputValues = PORTTC;
#endif
#if defined _STR91XF || defined _STM32
    ENABLE_LED_PORT();                                                   // ensure port is clocked and not in reset
    DEMO_LED_PORT &= ~DEMO_USER_PORTS;                                   // mask port bits
    DEMO_LED_PORT |= (ucPortOutputs & DEMO_USER_PORTS);                  // set initial output state
#endif
#if defined _LPC23XX || defined _LPC17XX
    unsigned long ulChange = ((ucPortOutputs & DEMO_USER_PORTS) << FIRST_USER_PORT);
    DEMO_LED_PORT_CLR = (~ulChange & (DEMO_USER_PORTS<<FIRST_USER_PORT));
    DEMO_LED_PORT_SET = (ulChange & (DEMO_USER_PORTS<<FIRST_USER_PORT));
    #ifdef _WINDOWS
    fnSimPorts();                                                        // ensure the simulator updates its port states
    #endif
#endif
#ifdef _LM3SXXXX
    if (iInitialisation) {
        ENABLE_LED_PORT();                                               // ensure port is clocked and not in reset
    }
    else {
        ucPortOutputs &= DEMO_LED_DDR;                                   // remove non-output bits
    }
    DEMO_LED_PORT = (DEMO_LED_PORT & ~DEMO_USER_PORTS) | (ucPortOutputs & DEMO_USER_PORTS); // {12} set port values
    temp_pars->temp_parameters.ucUserOutputValues = (unsigned char)DEMO_LED_PORT;
    #ifdef _WINDOWS
    fnPortChanges(0);                                                    // {14} ensure the simulator updates its port states
    #endif
#endif
}

// When this routine is called, the parameter block is set to a new valid block and the old block is deleted
//
extern int fnSaveNewPars(int iTemp)
{
#ifdef USE_PARAMETER_BLOCK
   // Network variables
   if (iTemp == SAVE_NEW_PARAMETERS_CHECK_CRITICAL) {                    // we check to see whether network parameters have been changed
       if (((temp_pars->temp_parameters.ucServers & ACTIVE_DHCP) != (parameters->ucServers & ACTIVE_DHCP)) || (uMemcmp((unsigned char *)&temp_pars->temp_network, (unsigned char *)&network_flash, sizeof(NETWORK_PARAMETERS)))) {
           return 1;                                                     // since a network parameter has been change, we warn the user that he/she will have to validate afterwards
       }
   }
    #ifndef USE_PAR_SWAP_BLOCK
   fnDelPar(INVALIDATE_PARAMETER_BLOCK);                                 // delete parameter block before continuing
   iTemp = SAVE_NEW_PARAMETERS;
    #endif
   fnSetPar((unsigned short)(PAR_NETWORK | TEMPORARY_PARAM_SET), (unsigned char *)&temp_pars->temp_network, sizeof(NETWORK_PARAMETERS)); // network parameters

   // device parameters
   fnSetPar((PAR_DEVICE | TEMPORARY_PARAM_SET), (unsigned char *)&temp_pars->temp_parameters, sizeof(PARS));  // copy the new set to the swap buffer (temp)

   if (iTemp != SAVE_NEW_PARAMETERS_VALIDATE) {
   #ifdef ACTIVE_FILE_SYSTEM
       fnDelPar(SWAP_PARAMETER_BLOCK);                                   // delete old block and validate new one.
   #endif
//     uMemcpy(&network, &temp_pars->temp_network, sizeof(NETWORK_PARAMETERS)); // update working parameter set - we never do this since we always reset after a network change (otherise we can have DHCP difficulties, since DHCP values will be overwritten)
       uMemcpy(parameters, &temp_pars->temp_parameters, sizeof(PARS));
   }
#endif
   return 0;
}


extern void fnRestoreFactory(void)
{
    fnSetDefaultNetwork(&temp_pars->temp_network);
    uMemcpy(&temp_pars->temp_network.ucOurMAC[0], &network.ucOurMAC[0], MAC_LENGTH); // return our mac since we never want to loose it
    uMemcpy(&temp_pars->temp_parameters, (unsigned char *)&cParameters, sizeof(PARS));
}

    #if defined USB_INTERFACE && defined ACTIVE_FILE_SYSTEM              // {8}
        typedef struct stUPLOAD_HEADER                                   // the start of a uTasker boot binary file
        {
            unsigned long  ulCodeLength;
            unsigned short usMagicNumber;
            unsigned short usCRC;
        } UPLOAD_HEADER;

        #ifdef SUB_FILE_SIZE
            #define SUBFILE_WRITE  , ucSubFileInProgress
            #define SUB_FILE_ON    ,SUB_FILE_TYPE
        #else
            #define SUBFILE_WRITE
            #define SUB_FILE_ON
        #endif

        #ifdef SUB_FILE_SIZE
            static unsigned char    ucSubFileInProgress = 0;
        #endif

        #if defined (_WINDOWS) || defined (_LITTLE_ENDIAN)
static void fnHeaderToLittleEndian(UPLOAD_HEADER *file_header)
{
    unsigned short usShort;
    unsigned long  ulLong;

    usShort = (file_header->usMagicNumber >> 8);
    file_header->usMagicNumber <<= 8;
    file_header->usMagicNumber |= usShort;

    ulLong = (file_header->ulCodeLength >> 24);
    ulLong |= (((file_header->ulCodeLength >> 16) & 0xff) << 8);
    ulLong |= (((file_header->ulCodeLength >> 8) & 0xff) << 16);
    ulLong |= (((file_header->ulCodeLength) & 0xff) << 24);
    file_header->ulCodeLength = ulLong;
}
        #endif



// Binary download support for USB
// This routine coordinates the task of opening the file on first contact, collecting buffers for saving and terminating when all data has been received.
// The format must be uTasker boot loader binary (which can also be encrypted)
extern void fnDownload(unsigned char *ptrData, QUEUE_TRANSFER Length)
{
    #define DOWNLOAD_STATE_IDLE       0
    #define DOWNLOAD_STATE_COLLECTING 1
    #define DOWNLOAD_STATE_SAVING     2
    #define DOWNLOAD_ERROR_STATE      3
    #define DOWNLOAD_BUFFER_SIZE      512

    static int iDownloadState = DOWNLOAD_STATE_IDLE;
    static unsigned long ulBufferData;
    static unsigned long ulExpectedData;
    static unsigned long ulSavedLength;
    static unsigned char *ptrWorkSpace = 0;
    static unsigned char *ptrBufferPointer;
    static unsigned char *ptrFile;
    switch (iDownloadState) {
    case 0:
        if (ptrWorkSpace == 0) {
            ptrWorkSpace = ptrBufferPointer = uMallocAlign(DOWNLOAD_BUFFER_SIZE, sizeof(unsigned long)); // get working memory on first contact (aligned to long word)
        }
        ulBufferData = 0;
        ulSavedLength = 0;
        ptrBufferPointer = ptrWorkSpace;
        iDownloadState = DOWNLOAD_STATE_COLLECTING;
        // fall through
    case DOWNLOAD_STATE_COLLECTING:
        while (ulBufferData < DOWNLOAD_BUFFER_SIZE) {
            if (Length == 0) {
                return;
            }
            *ptrBufferPointer++ = *ptrData++;
            Length--;
            ulBufferData++;
            if (ulBufferData == sizeof(UPLOAD_HEADER)) {                 // the boot loader header has been collected
                UPLOAD_HEADER *file_header = (UPLOAD_HEADER *)ptrWorkSpace;
#if defined (_WINDOWS) || defined (_LITTLE_ENDIAN)
                fnHeaderToLittleEndian(file_header);                     // convert header content to little endian format
#endif
                if (file_header->usMagicNumber != VALID_VERSION_MAGIC_NUMBER) {
                    fnDebugMsg("\r\nBAD FORMAT - quitting\r\n");
                    iDownloadState = DOWNLOAD_ERROR_STATE;
                    return;
                }
                if (file_header->ulCodeLength >= (SIZE_OF_FLASH/2)) {    // simple check that the length is valid
                    fnDebugMsg("\r\nBAD LENGTH - quitting\r\n");
                    iDownloadState = DOWNLOAD_ERROR_STATE;
                    return;
                }
                ulExpectedData = (file_header->ulCodeLength + sizeof(UPLOAD_HEADER)); // the length we will receive before terminating
                ptrFile = uOpenFile(USB_DOWNLOAD_FILE);                  // open a file to save the data to
        #ifdef SUB_FILE_SIZE
                ucSubFileInProgress = fnGetFileType(USB_DOWNLOAD_FILE);  // get file characteristics so that it is later handled correctly
        #endif
                iDownloadState = DOWNLOAD_STATE_SAVING;
                fnDebugMsg("Downloading..");
#if defined (_WINDOWS) || defined (_LITTLE_ENDIAN)
                fnHeaderToLittleEndian(file_header);                     // convert header back to original format
#endif
                break;
            }
        }
        // fall through
    case DOWNLOAD_STATE_SAVING:
        while (ulBufferData < DOWNLOAD_BUFFER_SIZE) {
            if (Length == 0) {
                break;
            }
            *ptrBufferPointer++ = *ptrData++;
            Length--;
            ulBufferData++;
            if (ulBufferData == DOWNLOAD_BUFFER_SIZE) {                  // a buffer has been collected so we save it
                fnDebugMsg("*");                                         // display progress
                uFileWrite(ptrFile, ptrWorkSpace, DOWNLOAD_BUFFER_SIZE SUBFILE_WRITE); // write the block to the file
                ulSavedLength += DOWNLOAD_BUFFER_SIZE;
                ulBufferData = 0;
                ptrBufferPointer = ptrWorkSpace;
            }
        }
        break;

    case DOWNLOAD_ERROR_STATE:
        if ((ptrData == 0) && (Length == 0)) {
            iDownloadState = 0;
        }
        return;
    }

    if ((ulSavedLength + ulBufferData) >= ulExpectedData) {              // if all data has been collected
        #ifdef SUPPORT_MIME_IDENTIFIER
        unsigned char ucMimeType = MIME_BINARY;
        #endif
        if ((ulExpectedData > ulSavedLength) != 0) {
            uFileWrite(ptrFile, ptrWorkSpace, (ulExpectedData - ulSavedLength) SUBFILE_WRITE); // write the remained to the file
        }
        #ifdef SUPPORT_MIME_IDENTIFIER                                   // close the file
        uFileCloseMime(ptrFile, &ucMimeType);                            // this will cause the file length and type to be written in the file
        #else
        uFileClose(ptrFile);                                             // this will cause the file length to be written in the file
        #endif
        fnDebugMsg("\r\nTerminated, resetting...");
        uTaskerMonoTimer( TASK_APPLICATION, (DELAY_LIMIT)(2*SEC), E_TIMER_SW_DELAYED_RESET );
    }
}

    #endif

    #ifdef TEST_CLIENT_SERVER

static int fnTestListener(USOCKET Socket, unsigned char ucEvent, unsigned char *ucIp_Data, unsigned short usPortLen)
{
  //static int iDelay = 0;
	switch (ucEvent) {
    case TCP_EVENT_CONNECTED:
      //fnTestTCPServer(START_SERVER_TEST);
      //return (fnTestTCPServer(TCP_SERVER_TX));
        fnDebugMsg("Connected\r\n");
        break;
    case TCP_EVENT_ACK:
        fnTestTCPServer(ALLOW_NEXT_MESSAGE); 
      //if (++iDelay > 5) {
            fnTestTCPServer(INCREMENT_SERVER_MESSAGE);        
            return (fnTestTCPServer(TCP_SERVER_TX));
      //}
        break;
    case TCP_EVENT_REGENERATE:
        return (fnTestTCPServer(TCP_SERVER_REPEAT));
    case TCP_EVENT_ABORT:
        fnDebugMsg("Connection reset\n\r");
    case TCP_EVENT_CLOSED:
        fnTestTCPServer(TCP_TEST_SERVER_LISTEN);
        break;
    case TCP_EVENT_DATA:
        fnTestTCPServer(INCREMENT_SERVER_MESSAGE);        
        return (fnTestTCPServer(TCP_SERVER_TX));
    case TCP_EVENT_CONREQ:
    case TCP_EVENT_PARTIAL_ACK:
    case TCP_EVENT_CLOSE:
        break;
    case TCP_EVENT_ARP_RESOLUTION_FAILED:
        fnDebugMsg("ARP failed\n\r");
        break;
    }
    return APP_ACCEPT;
}

#if defined _LPC23XX
static void server_int(void)
{
    fnInterruptMessage(OWN_TASK, TCP_SERVER_TEST);
}
#endif

static int fnTestTCPServer(int iAction)
{
    typedef struct stTCP_MESSAGE
    {
        TCP_HEADER     tTCP_Header;                                      // reserve header space
        unsigned char  ucTCP_Message[10];
    } TCP_MESSAGE;
    #define TEST_SERVERPORT    1234
    static const unsigned char ucRemoteIP[] = {192, 168, 0, 4};
    static int iTCPState = 0;
    static USOCKET test_socket = -1;
    static TCP_MESSAGE TCP_server_data;
    switch (iAction) {
    case TCP_TEST_SERVER_START:
        test_socket = fnGetTCP_Socket(TOS_MAXIMISE_THROUGHPUT, TCP_DEFAULT_TIMEOUT, fnTestListener);
    case TCP_TEST_SERVER_LISTEN:
        iTCPState = 0;
        fnTCP_Listen(test_socket, TEST_SERVERPORT, 0);
        break;
    case CONNECT_REMOTE:
        fnTCP_Connect(test_socket, (unsigned char *)ucRemoteIP, TEST_SERVERPORT, 0, 0);
        break;
    case START_SERVER_TEST:                                              // on connection from remote client
        {
        #if defined _LPC23XX
            static TIMER_INTERRUPT_SETUP timer_setup = {0};              // interrupt configuration parameters
            fnTestTCPServer(RESET_SERVER_MESSAGE);
            timer_setup.int_type = TIMER_INTERRUPT;
            timer_setup.int_priority = PRIORITY_TIMERS;
            timer_setup.int_handler = server_int;
            timer_setup.timer_reference = 2;                             // timer channel 2

            #ifdef _HW_TIMER_MODE
            timer_setup.timer_mode = TIMER_PERIODIC;                     // period timer interrupt        
            timer_setup.timer_value = TIMER_US_DELAY(9500);             // periodic timer us value
            #else
            timer_setup.timer_value = 0;
            timer_setup.timer_us_value = 9500;
            #endif
            fnConfigureInterrupt((void *)&timer_setup);                  // enter interrupt for server timer
        #endif
        }
        break;
    case ALLOW_NEXT_MESSAGE:
        iTCPState = 0;
        break;
    case INCREMENT_SERVER_MESSAGE:
        {
            int iIndex = (sizeof(TCP_server_data.ucTCP_Message) - 1);
            if (iTCPState != 0) {
                break;
            }
            while (iIndex >= 0) {
                if (++TCP_server_data.ucTCP_Message[iIndex] == 0) {
                    iIndex--;
                }
                break;
            }
        }
        break;
    case RESET_SERVER_MESSAGE:
        uMemset(&TCP_server_data.ucTCP_Message, 0, sizeof(TCP_server_data.ucTCP_Message));
        break;
    case TCP_SERVER_TX:
        if (iTCPState != 0) {
            break;
        }
    case TCP_SERVER_REPEAT:
        {
            static int iSent = 0;
            int iIndex = (sizeof(TCP_server_data.ucTCP_Message) - 1);
            if (TCP_server_data.ucTCP_Message[iIndex] == 1) {
                TCP_server_data.ucTCP_Message[iIndex] = 1;
            }
            if (iAction == TCP_SERVER_REPEAT) {
                iTCPState = 0;
                iSent--;
            }
            if (iSent > 1) {
                break;
            }
            if (fnSendTCP(test_socket, (unsigned char *)&TCP_server_data.tTCP_Header, sizeof(TCP_server_data.ucTCP_Message), TCP_FLAG_PUSH) > 0) {
                iTCPState = 1;                                               // we are expecting an ack
                iSent++;
                return APP_SENT_DATA;
            }
        }
        break;
    }
    return APP_ACCEPT;
}
    #endif
#else                                                                    // else not USE_MAINTENANCE {6}
    #ifdef SERIAL_INTERFACE
// Echo input or send start on password entry
//
extern void fnEchoInput(unsigned char *ucInputMessage, QUEUE_TRANSFER Length)
{
    fnWrite( SerialPortID, ucInputMessage, Length);                      // echo received character(s)
}
extern int fnInitiateLogin(unsigned short usNextState)
{
    return DIRECT_LOGIN;
}
extern int fnCommandInput(unsigned char *ptrData, unsigned short usLen, int iSource)
{
    return 0;
}
    #endif
#endif

#ifdef USE_TIME_SERVER
extern unsigned char fnAddTime(CHAR *ptrBuffer)                          // {7}
{
    unsigned long ulSeconds = ulPresentTime;
    unsigned char ucUnits = (unsigned char)(ulPresentTime/(60*60));

    ptrBuffer = fnBufferDec(ucUnits, (NO_TERMINATOR | LEADING_ZERO), ptrBuffer);
    *ptrBuffer++ = ':';
    ulSeconds -= (unsigned long)((unsigned long)ucUnits * 60 * 60);
    ucUnits = (unsigned char)(ulSeconds/60);
    ptrBuffer = fnBufferDec(ucUnits, (NO_TERMINATOR | LEADING_ZERO), ptrBuffer);
    *ptrBuffer++ = ':';
    ulSeconds -= (unsigned long)((unsigned long)ucUnits * 60);
    fnBufferDec(ulSeconds, (NO_TERMINATOR | LEADING_ZERO), ptrBuffer);

    return 8;                                                            // fixed length
}
#endif

#if defined _M5223X
// Determine the cause of the last reset
//
extern unsigned char fnAddResetCause(CHAR *ptrBuffer)
{
    #ifdef BACKUP_WATCHDOG_TIMER
    static const CHAR cWatchdog[]      = "WDOG";
    #endif
    static const CHAR cPowerOn[]       = "Power-on";
    static const CHAR cUndervoltage[]  = "Undervoltage";
    static const CHAR cClockLoss[]     = "Clock loss";
    static const CHAR cSoftwareReset[] = "Software";
    static const CHAR cExternalReset[] = "External";
    static const CHAR cUnknown[]       = "???";
    const CHAR *ptrStr;
    if (RESET_RSR == (RESET_POR | RESET_LVD)) {                          // power-on reset
        ptrStr = cPowerOn;
    }
    else if (RESET_RSR & RESET_LVD) {                                    // under voltage
        ptrStr = cUndervoltage;
    }
    #ifdef BACKUP_WATCHDOG_TIMER
    else if (RESET_RSR & (RESET_WDR | RESET_BWT)) {                      // background watchdog timer
        ptrStr = cWatchdog;
    }
    #endif
    else if (RESET_RSR & (RESET_LOL & RESET_LOC)) {                      // clock loss
        ptrStr = cClockLoss;
    }
    else if (RESET_RSR & RESET_SOFT) {                                   // commanded reset
        ptrStr = cSoftwareReset;
    }
    else if (RESET_RSR & RESET_EXT) {                                    // external reset
        ptrStr = cExternalReset;
    }
    else {                                                               // unexpected
        ptrStr = cUnknown;
    }
    return (uStrcpy(ptrBuffer, ptrStr) - ptrBuffer);                     // return the length of the string
}
#elif defined _LPC23XX
// Determine the cause of the last reset
//
extern unsigned char fnAddResetCause(CHAR *ptrBuffer)
{
    static unsigned char ucRSIR        = 0xff;
    static const CHAR cWatchdog[]      = "WDOG";
    static const CHAR cPowerOn[]       = "Power-on";
    static const CHAR cUndervoltage[]  = "Undervoltage";
    static const CHAR cExternalReset[] = "External";
    static const CHAR cUnknown[]       = "???";
    const CHAR *ptrStr;
    if (ucRSIR == 0xff) {
        ucRSIR = (unsigned char)RSIR;                                    // store the value for following requests
        RSIR = 0xff;                                                     // clear flags ready for next reset
    }
    if (ucRSIR & BODR) {                                                 // under voltage
        ptrStr = cUndervoltage;
    }
    else if (ucRSIR & WDTR) {                                            // watchdog timer
        ptrStr = cWatchdog;
    }

    else if (ucRSIR & EXTR) {                                            // external reset
        ptrStr = cExternalReset;
    }
    else if (ucRSIR & POR) {                                             // power-on reset
        ptrStr = cPowerOn;
    }
    else {                                                               // unexpected
        ptrStr = cUnknown;
    }
    return (uStrcpy(ptrBuffer, ptrStr) - ptrBuffer);                     // return the length of the string
}
#endif


extern CHAR *fnUpTime(CHAR *cValue)
{
    unsigned long ulUpTimeSecs = (uTaskerSystemTick / SEC);
    unsigned long ulUpTime     = ulUpTimeSecs /60/60/24;                 // the up time in days
    CHAR *cPtr;

    cPtr = fnBufferDec(ulUpTime, 0, cValue);
    *cPtr++ = ' ';
    *cPtr++ = 'D';
    *cPtr++ = 'a';
    *cPtr++ = 'y';
    *cPtr++ = 's';
    *cPtr++ = ' ';
    ulUpTimeSecs -= (ulUpTime * 60*60*24);                               // minus the days
    ulUpTime = (ulUpTimeSecs / (60*60));                                 // the UP time in hours
    cPtr = fnBufferDec(ulUpTime, 0, cPtr);
    *cPtr++ = ':';
    ulUpTimeSecs -= (ulUpTime * 60*60);                                  // minus the hours
    ulUpTime = (ulUpTimeSecs / (60));                                    // the UP time in minutes
    cPtr = fnBufferDec(ulUpTime, LEADING_ZERO, cPtr);
    *cPtr++ = ':';
    ulUpTimeSecs -= (ulUpTime * 60);                                     // minus the minutes
    cPtr = fnBufferDec(ulUpTimeSecs, LEADING_ZERO, cPtr);

    return cPtr;
}

extern CHAR *fnShowSN(CHAR *cValue)
{
    unsigned short usStart = temp_pars->temp_network.ucOurMAC[1];
    unsigned long  ulEnd   = temp_pars->temp_network.ucOurMAC[5];

    usStart += ((unsigned short)(temp_pars->temp_network.ucOurMAC[0]) << 8);

    ulEnd += ((unsigned long)(temp_pars->temp_network.ucOurMAC[4]) << 8);
    ulEnd += ((unsigned long)(temp_pars->temp_network.ucOurMAC[3]) << 16);
    ulEnd += ((unsigned long)(temp_pars->temp_network.ucOurMAC[2]) << 24);

    cValue = fnBufferDec(ulEnd,  LEADING_ZERO, cValue);
    *cValue++ = '-';
    cValue = fnBufferDec(usStart, (LEADING_ZERO | WITH_CR_LF), cValue);

    return (cValue);
}

extern void fnSetNewValue(int iType, CHAR *ptr_input)                    // {6}
{
    while ('+' == *ptr_input) {
        ptr_input++;
    }
    switch (iType) {
    case FLOW_CONTROL_HIGH_WATER:
        temp_pars->temp_parameters.ucFlowHigh = (unsigned char)fnDecStrHex(ptr_input);
        if (temp_pars->temp_parameters.ucFlowHigh > 99) {
            temp_pars->temp_parameters.ucFlowHigh = 99;
        }
        else if (temp_pars->temp_parameters.ucFlowHigh < 10) {
            temp_pars->temp_parameters.ucFlowHigh = 10;
        }
        break;

    case FLOW_CONTROL_LOW_WATER:
        temp_pars->temp_parameters.ucFlowLow = (unsigned char)fnDecStrHex(ptr_input);
        if (temp_pars->temp_parameters.ucFlowLow > 99) {
            temp_pars->temp_parameters.ucFlowLow = 99;
        }
        else if (temp_pars->temp_parameters.ucFlowLow < 10) {
            temp_pars->temp_parameters.ucFlowLow = 10;
        }
        break;

    case TELNET_PORT:
        temp_pars->temp_parameters.usTelnetPort = (unsigned short)fnDecStrHex(ptr_input);
        break;
    }
}

extern void fnConfigureFtpServer(unsigned short usTimeout)               // {3}{6}
{
#ifdef USE_FTP
    if (temp_pars->temp_parameters.ucServers & ACTIVE_FTP_SERVER) {      // should the FTP server be started?
        unsigned char ucFTP_mode = 0;
        if (temp_pars->temp_parameters.ucServers & ACTIVE_FTP_LOGIN) {
            ucFTP_mode = FTP_AUTHENTICATE;
        }
        fnStartFtp(usTimeout, ucFTP_mode);                               // start the FTP server
    }
    else {
        fnStopFtp();
    }
#endif
}


// We support the network transmit routine here
//
extern QUEUE_TRANSFER fnNetworkTx(unsigned char *output_buffer, QUEUE_TRANSFER nr_of_bytes)
{
#if defined USE_TELNET && defined USE_MAINTENANCE                        // {6} no Telnet without USE_MAINTENANCE
    if (output_buffer == 0) {                                            // {18} if a check of output buffer size call the correct function
        return (fnSendBufTCP(Telnet_socket, (unsigned char *)&OurTask, nr_of_bytes, TCP_BUF_CHECK));
    }
    return (fnSendBufTCP(Telnet_socket, output_buffer, nr_of_bytes, (TCP_BUF_SEND | TCP_BUF_SEND_REPORT_COPY)));
#else
    return 0;
#endif
}


